"""Python implementations of Scenic language constructs.

This module is automatically imported by all Scenic programs. In addition to
defining the built-in functions, operators, specifiers, etc., it also stores
global state such as the list of all created Scenic objects.

.. highlight:: scenic-grammar
"""

__all__ = (
    # Primitive statements and functions
    "ego",
    "workspace",
    "new",
    "require",
    "resample",
    "param",
    "globalParameters",
    "mutate",
    "verbosePrint",
    "localPath",
    "model",
    "simulator",
    "simulation",
    "require_monitor",
    "terminate_when",
    "terminate_simulation_when",
    "terminate_after",
    "in_initial_scenario",
    "override",
    "record",
    "record_initial",
    "record_final",
    "sin",
    "cos",
    "hypot",
    "max",
    "min",
    "filter",
    "str",
    "float",
    "int",
    "round",
    "len",
    "range",
    # Prefix operators
    "Visible",
    "NotVisible",
    "Front",
    "Back",
    "Left",
    "Right",
    "FrontLeft",
    "FrontRight",
    "BackLeft",
    "BackRight",
    "Top",
    "Bottom",
    "TopFrontLeft",
    "TopFrontRight",
    "TopBackLeft",
    "TopBackRight",
    "BottomFrontLeft",
    "BottomFrontRight",
    "BottomBackLeft",
    "BottomBackRight",
    "RelativeHeading",
    "ApparentHeading",
    "RelativePosition",
    "DistanceFrom",
    "DistancePast",
    "Follow",
    "AngleTo",
    "AngleFrom",
    "AltitudeTo",
    "AltitudeFrom",
    # Infix operators
    "FieldAt",
    "RelativeTo",
    "OffsetAlong",
    "CanSee",
    "Intersects",
    "Until",
    "Implies",
    "VisibleFromOp",
    "NotVisibleFromOp",
    # Primitive types
    "Vector",
    "Orientation",
    "VectorField",
    "PolygonalVectorField",
    "Shape",
    "MeshShape",
    "BoxShape",
    "CylinderShape",
    "ConeShape",
    "SpheroidShape",
    "MeshVolumeRegion",
    "MeshSurfaceRegion",
    "BoxRegion",
    "SpheroidRegion",
    "PathRegion",
    "Region",
    "PointSetRegion",
    "RectangularRegion",
    "CircularRegion",
    "SectorRegion",
    "PolygonalRegion",
    "PolylineRegion",
    "Workspace",
    "Mutator",
    "Range",
    "DiscreteRange",
    "Options",
    "Uniform",
    "Discrete",
    "Normal",
    "TruncatedNormal",
    "VerifaiParameter",
    "VerifaiRange",
    "VerifaiDiscreteRange",
    "VerifaiOptions",
    # Constructible types
    "Point",
    "OrientedPoint",
    "Object",
    # Specifiers
    "With",
    "At",
    "In",
    "ContainedIn",
    "On",
    "Beyond",
    "VisibleFrom",
    "NotVisibleFrom",
    "VisibleSpec",
    "NotVisibleSpec",
    "OffsetBy",
    "OffsetAlongSpec",
    "Facing",
    "ApparentlyFacing",
    "FacingToward",
    "FacingDirectlyToward",
    "FacingAwayFrom",
    "FacingDirectlyAwayFrom",
    "LeftSpec",
    "RightSpec",
    "Ahead",
    "Behind",
    "Above",
    "Below",
    "Following",
    # Constants
    "everywhere",
    "nowhere",
    # Exceptions
    "GuardViolation",
    "PreconditionViolation",
    "InvariantViolation",
    "RejectionException",
    # Internal APIs     # TODO remove?
    "_scenic_default",
    "Behavior",
    "Monitor",
    "_makeTerminationAction",
    "_makeSimulationTerminationAction",
    "BlockConclusion",
    "runTryInterrupt",
    "wrapStarredValue",
    "callWithStarArgs",
    "Modifier",
    "DynamicScenario",
    # Proposition Factories
    "AtomicProposition",
    "PropositionAnd",
    "PropositionOr",
    "PropositionNot",
    "Always",
    "Eventually",
    "Next",
)

# various Python types and functions used in the language but defined elsewhere
from scenic.core.distributions import (
    DiscreteRange,
    Normal,
    Options,
    RandomControlFlowError,
    Range,
    TruncatedNormal,
    Uniform,
)
from scenic.core.dynamics.behaviors import Behavior, Monitor
from scenic.core.dynamics.guards import (
    GuardViolation,
    InvariantViolation,
    PreconditionViolation,
)
from scenic.core.dynamics.invocables import BlockConclusion, runTryInterrupt
from scenic.core.dynamics.scenarios import DynamicScenario
from scenic.core.external_params import (
    VerifaiDiscreteRange,
    VerifaiOptions,
    VerifaiParameter,
    VerifaiRange,
)
from scenic.core.geometry import cos, hypot, max, min, sin
from scenic.core.object_types import Mutator, Object, OrientedPoint, Point
from scenic.core.regions import (
    BoxRegion,
    CircularRegion,
    MeshSurfaceRegion,
    MeshVolumeRegion,
    PathRegion,
    PointSetRegion,
    PolygonalRegion,
    PolylineRegion,
    RectangularRegion,
    Region,
    SectorRegion,
    SpheroidRegion,
    everywhere,
    nowhere,
)
from scenic.core.shapes import (
    BoxShape,
    ConeShape,
    CylinderShape,
    MeshShape,
    Shape,
    SpheroidShape,
)
from scenic.core.specifiers import PropertyDefault as _scenic_default
from scenic.core.vectors import PolygonalVectorField, Vector, VectorField
from scenic.core.workspaces import Workspace

Discrete = Options

# isort: split

# everything that should not be directly accessible from the language is imported here:
import builtins
import collections.abc
from contextlib import contextmanager
import functools
import importlib
import numbers
from pathlib import Path
import sys
import traceback
import typing

from scenic.core.distributions import (
    Distribution,
    MultiplexerDistribution,
    RejectionException,
    StarredDistribution,
    TupleDistribution,
    canUnpackDistributions,
    distributionFunction,
    needsSampling,
    toDistribution,
)
from scenic.core.dynamics.actions import _EndScenarioAction, _EndSimulationAction
import scenic.core.errors as errors
from scenic.core.errors import InvalidScenarioError, ScenicSyntaxError
from scenic.core.external_params import ExternalParameter
from scenic.core.geometry import apparentHeadingAtPoint, normalizeAngle
from scenic.core.lazy_eval import (
    DelayedArgument,
    isLazy,
    needsLazyEvaluation,
    requiredProperties,
    valueInContext,
)
import scenic.core.object_types
from scenic.core.object_types import Constructible, Object2D, OrientedPoint2D, Point2D
import scenic.core.propositions as propositions
from scenic.core.regions import convertToFootprint
import scenic.core.requirements as requirements
from scenic.core.simulators import RejectSimulationException
from scenic.core.specifiers import ModifyingSpecifier, Specifier
from scenic.core.type_support import (
    Heading,
    canCoerce,
    coerce,
    evaluateRequiringEqualTypes,
    isA,
    toHeading,
    toOrientation,
    toScalar,
    toType,
    toTypes,
    toVector,
    underlyingType,
)
from scenic.core.vectors import Orientation, alwaysGlobalOrientation

### Internals

activity = 0
currentScenario = None
scenarioStack = []
scenarios = []
evaluatingRequirement = False
_globalParameters = {}
lockedParameters = set()
lockedModel = None
loadingModel = False
currentSimulation = None
inInitialScenario = True
runningScenarios = []  # in order, oldest first
currentBehavior = None
simulatorFactory = None
evaluatingGuard = False
mode2D = False
_originalConstructibles = (Point, OrientedPoint, Object)
BUFFERING_PITCH = 0.1

## APIs used internally by the rest of Scenic

# Scenic compilation


def isActive():
    """Are we in the middle of compiling a Scenic module?

    The 'activity' global can be >1 when Scenic modules in turn import other
    Scenic modules.
    """
    return activity > 0


def activate(options, namespace=None):
    """Activate the veneer when beginning to compile a Scenic module."""
    global activity, _globalParameters, lockedParameters, lockedModel, currentScenario
    if options.paramOverrides or options.modelOverride:
        assert activity == 0
        _globalParameters.update(options.paramOverrides)
        lockedParameters = set(options.paramOverrides)
        lockedModel = options.modelOverride

    # If we are in 2D mode, set the global flag and replace all classes
    # with their 2D compatibility counterparts.
    if options.mode2D:
        global mode2D, Point, OrientedPoint, Object
        assert mode2D or activity == 0
        mode2D = True
        Point = Point2D
        OrientedPoint = OrientedPoint2D
        Object = Object2D
        scenic.core.object_types.Point = Point
        scenic.core.object_types.OrientedPoint = OrientedPoint
        scenic.core.object_types.Object = Object

    activity += 1
    assert not evaluatingRequirement
    assert not evaluatingGuard
    assert currentSimulation is None
    # placeholder scenario for top-level code
    newScenario = DynamicScenario._dummy(namespace)
    scenarioStack.append(newScenario)
    currentScenario = newScenario


def deactivate():
    """Deactivate the veneer after compiling a Scenic module."""
    global activity, _globalParameters, lockedParameters, lockedModel, mode2D
    global currentScenario, scenarios, scenarioStack, simulatorFactory
    activity -= 1
    assert activity >= 0
    assert not evaluatingRequirement
    assert not evaluatingGuard
    assert currentSimulation is None
    scenarioStack.pop()
    assert len(scenarioStack) == activity
    scenarios = []

    if activity == 0:
        lockedParameters = set()
        lockedModel = None
        currentScenario = None
        simulatorFactory = None
        _globalParameters = {}

        if mode2D:
            global Point, OrientedPoint, Object
            mode2D = False
            Point, OrientedPoint, Object = _originalConstructibles
            scenic.core.object_types.Point = Point
            scenic.core.object_types.OrientedPoint = OrientedPoint
            scenic.core.object_types.Object = Object
    else:
        currentScenario = scenarioStack[-1]


# Instance/Object creation


def registerInstance(inst):
    """Add a Scenic instance to the global list of created objects.

    This is called by the Point/OrientedPoint constructor.
    """
    if currentScenario:
        assert isinstance(inst, Constructible)
        currentScenario._registerInstance(inst)


def registerObject(obj):
    """Add a Scenic object to the global list of created objects.

    This is called by the Object constructor.
    """
    if evaluatingRequirement:
        raise InvalidScenarioError("tried to create an object inside a requirement")
    elif currentBehavior is not None:
        raise InvalidScenarioError("tried to create an object inside a behavior")
    elif activity > 0 or currentScenario:
        assert not evaluatingRequirement
        assert isinstance(obj, Object)
        currentScenario._registerObject(obj)
        if currentSimulation:
            currentSimulation._createObject(obj)


# External parameter creation


def registerExternalParameter(value):
    """Register a parameter whose value is given by an external sampler."""
    if activity > 0:
        assert isinstance(value, ExternalParameter)
        currentScenario._externalParameters.append(value)


# Function call support


def wrapStarredValue(value, lineno):
    if isinstance(value, TupleDistribution) or not needsSampling(value):
        return value
    elif isinstance(value, Distribution):
        return [StarredDistribution(value, lineno)]
    else:
        raise TypeError(f"iterable unpacking cannot be applied to {value}")


def callWithStarArgs(_func_to_call, *args, **kwargs):
    if not canUnpackDistributions(_func_to_call):
        # wrap function to delay evaluation until starred distributions are sampled
        _func_to_call = distributionFunction(_func_to_call)
    return _func_to_call(*args, **kwargs)


# Simulations


def instantiateSimulator(factory, params):
    global _globalParameters
    assert not _globalParameters  # TODO improve hack?
    _globalParameters = dict(params)
    try:
        return factory()
    finally:
        _globalParameters = {}


def beginSimulation(sim):
    global currentSimulation, currentScenario, inInitialScenario, runningScenarios
    global _globalParameters
    if isActive():
        raise RuntimeError("tried to start simulation during Scenic compilation!")
    assert currentSimulation is None
    assert currentScenario is None
    assert not scenarioStack
    currentSimulation = sim
    currentScenario = sim.scene.dynamicScenario
    runningScenarios = []  # will be updated by DynamicScenario._start
    inInitialScenario = currentScenario._setup is None
    currentScenario._bindTo(sim.scene)
    _globalParameters = dict(sim.scene.params)

    # rebind globals that could be referenced by behaviors to their sampled values
    for modName, (
        namespace,
        sampledNS,
        originalNS,
    ) in sim.scene.behaviorNamespaces.items():
        namespace.clear()
        namespace.update(sampledNS)


def endSimulation(sim):
    global currentSimulation, currentScenario, currentBehavior, runningScenarios
    global _globalParameters
    currentSimulation = None
    currentScenario = None
    runningScenarios = []
    currentBehavior = None
    _globalParameters = {}

    for modName, (
        namespace,
        sampledNS,
        originalNS,
    ) in sim.scene.behaviorNamespaces.items():
        namespace.clear()
        namespace.update(originalNS)


def simulationInProgress():
    return currentSimulation is not None


# Requirements


@contextmanager
def executeInRequirement(scenario, boundEgo, values):
    global evaluatingRequirement, currentScenario
    assert activity == 0
    assert not evaluatingRequirement
    evaluatingRequirement = True
    if currentScenario is None:
        currentScenario = scenario
        clearScenario = True
    else:
        assert currentScenario is scenario
        clearScenario = False
    oldEgo = currentScenario._ego
    oldObjects = currentScenario._objects

    currentScenario._objects = tuple(values[obj] for obj in currentScenario.objects)

    if boundEgo:
        currentScenario._ego = boundEgo
    try:
        yield
    except RandomControlFlowError as e:
        # Such errors should not be possible inside a requirement, since all values
        # should have already been sampled: something's gone wrong with our rebinding.
        raise RuntimeError("internal error: requirement dependency not sampled") from e
    finally:
        evaluatingRequirement = False
        currentScenario._ego = oldEgo
        currentScenario._objects = oldObjects
        if clearScenario:
            currentScenario = None


# Dynamic scenarios


def registerDynamicScenarioClass(cls):
    scenarios.append(cls)


@contextmanager
def executeInScenario(scenario, inheritEgo=False):
    global currentScenario
    oldScenario = currentScenario
    if inheritEgo and oldScenario is not None:
        scenario._ego = oldScenario._ego  # inherit ego from parent
    currentScenario = scenario
    try:
        yield
    except AttributeError as e:
        # Convert confusing AttributeErrors from trying to access nonexistent scenario
        # variables into NameErrors, which is what the user would expect. The information
        # needed to do this was made available in Python 3.10, but unfortunately could be
        # wrong until 3.10.3: see bpo-46940.
        if sys.version_info >= (3, 10, 3) and isinstance(e.obj, DynamicScenario):
            newExc = NameError(f"name '{e.name}' is not defined", name=e.name)
            raise newExc.with_traceback(e.__traceback__)
        else:
            raise
    finally:
        currentScenario = oldScenario


def prepareScenario(scenario):
    if currentSimulation:
        verbosePrint(f"Starting scenario {scenario}", level=3)


def finishScenarioSetup(scenario):
    global inInitialScenario
    inInitialScenario = False


def startScenario(scenario):
    assert scenario not in runningScenarios
    runningScenarios.append(scenario)


def endScenario(scenario, reason, quiet=False):
    runningScenarios.remove(scenario)
    if not quiet:
        verbosePrint(f"Stopping scenario {scenario} because: {reason}", level=3)


# Dynamic behaviors


@contextmanager
def executeInBehavior(behavior):
    global currentBehavior
    oldBehavior = currentBehavior
    currentBehavior = behavior
    try:
        yield
    except AttributeError as e:
        # See comment for corresponding code in executeInScenario
        if sys.version_info >= (3, 10, 3) and isinstance(e.obj, Behavior):
            newExc = NameError(f"name '{e.name}' is not defined", name=e.name)
            raise newExc.with_traceback(e.__traceback__)
        else:
            raise
    finally:
        currentBehavior = oldBehavior


@contextmanager
def executeInGuard():
    global evaluatingGuard
    assert not evaluatingGuard
    evaluatingGuard = True
    try:
        yield
    finally:
        evaluatingGuard = False


def _makeTerminationAction(agent, line):
    assert activity == 0
    if agent:
        scenario = agent._parentScenario()
        assert scenario is not None
    else:
        scenario = None
    return _EndScenarioAction(scenario, line)


def _makeSimulationTerminationAction(line):
    assert activity == 0
    return _EndSimulationAction(line)


### Parsing support


class Modifier(typing.NamedTuple):
    name: str
    value: typing.Any
    terminator: typing.Optional[str] = None


### Primitive statements and functions


def new(cls, specifiers):
    if not (isinstance(cls, type) and issubclass(cls, Constructible)):
        raise TypeError(f'"{cls.__name__}" is not a Scenic class')
    return cls._withSpecifiers(specifiers)


def ego(obj=None):
    """Function implementing loads and stores to the 'ego' pseudo-variable.

    The translator calls this with no arguments for loads, and with the source
    value for stores.
    """
    egoObject = currentScenario._ego
    if obj is None:
        if egoObject is None:
            raise InvalidScenarioError("referred to ego object not yet assigned")
    elif not isinstance(obj, Object):
        if isinstance(obj, type) and issubclass(obj, Object):
            suffix = " (perhaps you forgot 'new'?)"
        else:
            suffix = ""
        ty = type(obj).__name__
        raise TypeError(f"tried to make non-object (of type {ty}) the ego object{suffix}")
    else:
        currentScenario._ego = obj
        for scenario in runningScenarios:
            if scenario._ego is None:
                scenario._ego = obj
    return egoObject


def workspace(workspace=None):
    """Function implementing loads and stores to the 'workspace' pseudo-variable.

    See `ego`.
    """
    if workspace is None:
        if currentScenario._workspace is None:
            raise InvalidScenarioError("referred to workspace not yet assigned")
    elif not isinstance(workspace, Workspace):
        raise TypeError(f"workspace {workspace} is not a Workspace")
    elif needsSampling(workspace):
        raise InvalidScenarioError("workspace must be a fixed region")
    elif needsLazyEvaluation(workspace):
        raise InvalidScenarioError(
            "workspace uses value undefined " "outside of object definition"
        )
    else:
        currentScenario._workspace = workspace
    return currentScenario._workspace


def require(reqID, req, line, name, prob=1):
    """Function implementing the require statement."""
    if not name:
        name = f"requirement on line {line}"
    if evaluatingRequirement:
        raise InvalidScenarioError("tried to create a requirement inside a requirement")
    if req.has_temporal_operator and prob != 1:
        raise InvalidScenarioError(
            "requirements with temporal operators must have probability of 1"
        )
    if currentSimulation is not None:  # requirement being evaluated at runtime
        if req.has_temporal_operator:
            # support monitors on dynamic requirements and create dynamic requirements
            currentScenario._addDynamicRequirement(
                requirements.RequirementType.require, req, line, name
            )
        else:
            if prob >= 1 or Range(0, 1) <= prob:  # use Range so value can be recorded
                result = req.evaluate()
                assert not needsSampling(result)
                if needsLazyEvaluation(result):
                    raise InvalidScenarioError(
                        f"requirement on line {line} uses value"
                        " undefined outside of object definition"
                    )
                if not result:
                    raise RejectSimulationException(name)
    else:  # requirement being defined at compile time
        currentScenario._addRequirement(
            requirements.RequirementType.require, reqID, req, line, name, prob
        )


def require_monitor(reqID, value, line, name):
    if not name:
        name = f"requirement on line {line}"
    if currentSimulation is not None:
        monitor = value.evaluate()
        assert not needsSampling(monitor)
        if needsLazyEvaluation(monitor):
            raise InvalidScenarioError(
                f"requirement on line {line} uses value"
                " undefined outside of object definition"
            )
        if not isinstance(monitor, Monitor):
            raise TypeError(f'"require monitor X" with X not a monitor on line {line}')
        currentScenario._addMonitor(monitor)
    else:
        currentScenario._addRequirement(
            requirements.RequirementType.monitor, reqID, value, line, name, 1
        )


def record(reqID, value, line, name):
    if not name:
        name = f"record{line}"
    makeRequirement(requirements.RequirementType.record, reqID, value, line, name)


def record_initial(reqID, value, line, name):
    if not name:
        name = f"record{line}"
    makeRequirement(requirements.RequirementType.recordInitial, reqID, value, line, name)


def record_final(reqID, value, line, name):
    if not name:
        name = f"record{line}"
    makeRequirement(requirements.RequirementType.recordFinal, reqID, value, line, name)


def require_always(reqID, req, line, name):
    """Function implementing the 'require always' statement."""
    if not name:
        name = f"requirement on line {line}"
    makeRequirement(requirements.RequirementType.requireAlways, reqID, req, line, name)


def require_eventually(reqID, req, line, name):
    """Function implementing the 'require eventually' statement."""
    if not name:
        name = f"requirement on line {line}"
    makeRequirement(
        requirements.RequirementType.requireEventually, reqID, req, line, name
    )


def terminate_when(reqID, req, line, name):
    """Function implementing the 'terminate when' statement."""
    if not name:
        name = f"termination condition on line {line}"
    makeRequirement(requirements.RequirementType.terminateWhen, reqID, req, line, name)


def terminate_simulation_when(reqID, req, line, name):
    """Function implementing the 'terminate simulation when' statement."""
    if not name:
        name = f"termination condition on line {line}"
    makeRequirement(
        requirements.RequirementType.terminateSimulationWhen, reqID, req, line, name
    )


def makeRequirement(ty, reqID, req, line, name):
    if evaluatingRequirement:
        raise InvalidScenarioError(f'tried to use "{ty.value}" inside a requirement')
    elif currentBehavior is not None:
        raise InvalidScenarioError(f'"{ty.value}" inside a behavior on line {line}')
    elif currentSimulation is not None:
        currentScenario._addDynamicRequirement(ty, req, line, name)
    else:  # requirement being defined at compile time
        currentScenario._addRequirement(ty, reqID, req, line, name, 1)


def terminate_after(timeLimit, terminator=None):
    if not isinstance(timeLimit, (builtins.float, builtins.int)):
        raise TypeError('"terminate after N" with N not a number')
    assert terminator in (None, "seconds", "steps")
    inSeconds = terminator != "steps"
    currentScenario._setTimeLimit(timeLimit, inSeconds=inSeconds)


def resample(dist):
    """The built-in resample function."""
    if not isinstance(dist, Distribution):
        return dist
    try:
        return dist.clone()
    except NotImplementedError:
        raise TypeError("cannot resample non-primitive distribution") from None


def verbosePrint(
    *objects, level=1, indent=True, sep=" ", end="\n", file=sys.stdout, flush=False
):
    """Built-in function printing a message only in verbose mode.

    Scenic's verbosity may be set using the :option:`-v` command-line option.
    The simplest way to use this function is with code like
    :scenic:`verbosePrint('hello world!')` or :scenic:`verbosePrint('details here', level=3)`;
    the other keyword arguments are probably only useful when replacing more complex uses
    of the Python `print` function.

    Args:
        objects: Object(s) to print (`str` will be called to make them strings).
        level (int): Minimum verbosity level at which to print. Default is 1.
        indent (bool): Whether to indent the message to align with messages generated by
            Scenic (default true).
        sep, end, file, flush: As in `print`.
    """
    if errors.verbosityLevel >= level:
        if indent:
            if currentSimulation:
                indent = "      " if errors.verbosityLevel >= 3 else "  "
            else:
                indent = "  " * activity if errors.verbosityLevel >= 2 else "  "
            print(indent, end="", file=file)
        print(*objects, sep=sep, end=end, file=file, flush=flush)


def localPath(relpath):
    """Convert a path relative to the calling Scenic file into an absolute path.

    For example, :scenic:`localPath('resource.dat')` evaluates to the absolute path
    of a file called ``resource.dat`` located in the same directory as the
    Scenic file where this expression appears. Note that the path is returned as a
    `pathlib.Path` object.
    """
    filename = traceback.extract_stack(limit=2)[0].filename
    base = Path(filename).parent
    return base.joinpath(relpath).resolve()


def simulation():
    """Get the currently-running `Simulation`.

    May only be called from code that runs at simulation time, e.g. inside
    :term:`dynamic behaviors` and :keyword:`compose` blocks of scenarios.
    """
    if isActive():
        raise InvalidScenarioError("used simulation() outside a behavior")
    assert currentSimulation is not None
    return currentSimulation


def simulator(sim):
    global simulatorFactory
    simulatorFactory = sim


def in_initial_scenario():
    return inInitialScenario


def override(*args):
    if len(args) < 1:
        raise TypeError('"override" missing an object')
    elif len(args) < 2:
        raise TypeError('"override" missing a list of specifiers')
    obj = args[0]
    if not isinstance(obj, Object):
        raise TypeError(f'"override" passed non-Object {obj}')
    specs = args[1:]
    for spec in specs:
        assert isinstance(spec, Specifier), spec

    currentScenario._override(obj, specs)


def model(namespace, modelName):
    global loadingModel
    if loadingModel:
        raise InvalidScenarioError('Scenic world model itself uses the "model" statement')
    if lockedModel is not None:
        modelName = lockedModel
    try:
        loadingModel = True
        module = importlib.import_module(modelName)
    except ModuleNotFoundError as e:
        if e.name == modelName:
            raise InvalidScenarioError(
                f"could not import world model {modelName}"
            ) from None
        else:
            raise
    finally:
        loadingModel = False
    names = module.__dict__.get("__all__", None)
    if names is not None:
        for name in names:
            namespace[name] = getattr(module, name)
    else:
        for name, value in module.__dict__.items():
            if not name.startswith("_"):
                namespace[name] = value


def param(params):
    """Function implementing the param statement."""
    global loadingModel
    if evaluatingRequirement:
        raise InvalidScenarioError(
            "tried to create a global parameter inside a requirement"
        )
    elif currentSimulation is not None:
        raise InvalidScenarioError(
            "tried to create a global parameter during a simulation"
        )
    for name, value in params.items():
        if name not in lockedParameters and (
            not loadingModel or name not in _globalParameters
        ):
            _globalParameters[name] = toDistribution(value)


class ParameterTableProxy(collections.abc.Mapping):
    def __init__(self, map):
        object.__setattr__(self, "_internal_map", map)

    def __getitem__(self, name):
        return self._internal_map[name]

    def __iter__(self):
        return iter(self._internal_map)

    def __len__(self):
        return len(self._internal_map)

    def __getattr__(self, name):
        return self.__getitem__(name)  # allow namedtuple-like access

    def __setattr__(self, name, value):
        raise InvalidScenarioError(
            'cannot modify globalParameters (use "param" statement)'
        )

    def _clone_table(self):
        return ParameterTableProxy(self._internal_map.copy())


def globalParameters():
    return ParameterTableProxy(_globalParameters)


def mutate(*objects, scale=1):
    """Function implementing the mutate statement."""
    if evaluatingRequirement:
        raise InvalidScenarioError("used mutate statement inside a requirement")
    if len(objects) == 0:
        objects = currentScenario._objects
    if not isinstance(scale, (builtins.int, builtins.float)):
        raise TypeError('"mutate X by Y" with Y not a number')
    for obj in objects:
        if not isinstance(obj, Object):
            raise TypeError('"mutate X" with X not an object')
        obj.mutationScale = scale
        # Object will now require sampling even if it has no explicit dependencies.
        obj._needsSampling = True
        obj._isLazy = True


### Prefix operators


def Visible(region):
    """The :grammar:`visible <region>` operator."""
    region = toType(region, Region, '"visible X" with X not a Region')
    return region.intersect(ego().visibleRegion)


def NotVisible(region):
    """The :grammar:`not visible <region>` operator."""
    region = toType(region, Region, '"not visible X" with X not a Region')
    return region.difference(ego().visibleRegion)


# front of <object>, etc.
ops = (
    "front",
    "back",
    "left",
    "right",
    "front left",
    "front right",
    "back left",
    "back right",
    "top",
    "bottom",
    "top front left",
    "top front right",
    "top back left",
    "top back right",
    "bottom front left",
    "bottom front right",
    "bottom back left",
    "bottom back right",
)
template = '''\
def {function}(X):
    """The :grammar:`{syntax} of <object>` operator."""
    if not isinstance(X, Object):
        raise TypeError('"{syntax} of X" with X not an Object')
    return X.{property}
'''
for op in ops:
    func = "".join(word.capitalize() for word in op.split(" "))
    prop = func[0].lower() + func[1:]
    definition = template.format(function=func, syntax=op, property=prop)
    exec(definition)

### Infix operators


def FieldAt(X, Y):
    """The :grammar:`<vector field> at <vector>` operator."""
    if isinstance(X, type) and issubclass(X, Constructible):
        raise TypeError('"X at Y" with X not a vector field. (Perhaps you forgot "new"?)')

    if not isA(X, VectorField):
        raise TypeError('"X at Y" with X not a vector field')
    Y = toVector(Y, '"X at Y" with Y not a vector')
    return X[Y]


def RelativeTo(X, Y) -> typing.Union[Vector, builtins.float, Orientation]:
    """The :scenic:`{X} relative to {Y}` polymorphic operator.

    Allowed forms::

        <value> relative to <value> # with at least one a field, the other a field or heading
        <vector> relative to <oriented point> # and vice versa
        <vector> relative to <vector>
        <heading> relative to <heading>
        <orientation> relative to <orientation>
    """

    # Define lazy RelativeTo helper
    @distributionFunction
    def lazyRelativeTo(X, Y) -> typing.Union[Vector, builtins.float, Orientation]:
        return RelativeTo(X, Y)

    # Define type helpers
    def knownOrientation(thing):
        return isA(thing, Orientation) or (
            (not isLazy(thing))
            and canCoerce(thing, Orientation)
            and (not canCoerce(thing, Vector))
        )

    def knownHeading(thing):
        return isA(thing, numbers.Real) or (
            (not isLazy(thing)) and canCoerce(thing, Heading)
        )

    def knownVector(thing):
        return isA(thing, Vector) or ((not isLazy(thing)) and canCoerce(thing, Vector))

    xf, yf = isA(X, VectorField), isA(Y, VectorField)
    if xf or yf:
        if xf and yf and X.valueType != Y.valueType:
            raise TypeError('"X relative to Y" with X, Y fields of different types')
        fieldType = X.valueType if xf else Y.valueType
        error = '"X relative to Y" with field and value of different types'

        def helper(context):
            pos = context.position.toVector()
            xp = X[pos] if xf else toType(X, fieldType, error)
            yp = Y[pos] if yf else toType(Y, fieldType, error)
            return yp + xp

        return DelayedArgument({"position"}, helper)

    elif isA(X, OrientedPoint) or isA(Y, OrientedPoint):
        # Ensure X and Y aren't both oriented points
        if isA(X, OrientedPoint) and isA(Y, OrientedPoint):
            raise TypeError('"X relative to Y" with X, Y both oriented points')

        # Extract the single oriented point and the other value
        if isA(X, OrientedPoint):
            op = X
            other = Y
        else:
            op = Y
            other = X

        # Check the other value's type
        if isA(other, numbers.Real):
            return op.heading + toHeading(other)
        elif isA(other, Orientation):
            return toOrientation(Y) * toOrientation(X)
        elif knownVector(other):
            other = toVector(other)
            return op.relativize(other)

        # This case doesn't match (for now at least). Fall through.
        pass

    elif knownOrientation(X) and knownOrientation(Y):
        xf = toOrientation(X)
        yf = toOrientation(Y)

        return yf * xf

    elif knownHeading(X) and knownHeading(Y):
        xf = toHeading(X, f'"X relative to Y" with Y a heading but X a {type(X)}')
        yf = toHeading(Y, f'"X relative to Y" with X a heading but Y a {type(Y)}')

        return xf + yf

    elif knownVector(X) or knownVector(Y):
        xf = toVector(X, f'"X relative to Y" with Y a vector but X a {type(X)}')
        yf = toVector(Y, f'"X relative to Y" with X a vector but Y a {type(Y)}')

        return xf + yf

    if isLazy(X) or isLazy(Y):
        # We can't determine what case to use at this point. Try again when things are sampled.
        return lazyRelativeTo(X, Y)

    raise TypeError(
        f'"X relative to Y" with X and Y incompatible types (X a {type(X)}, Y a {type(Y)})'
    )


def OffsetAlong(X, H, Y):
    """The :scenic:`{X} offset along {H} by {Y}` polymorphic operator.

    Allowed forms::

        <vector> offset along <heading> by <vector>
        <vector> offset along <field> by <vector>
    """
    X = toVector(X, '"X offset along H by Y" with X not a vector')
    Y = toVector(Y, '"X offset along H by Y" with Y not a vector')
    if isA(H, VectorField):
        H = H[X]
    H = toOrientation(
        H, '"X offset along H by Y" with H not an orientation or vector field'
    )
    return X.offsetLocally(H, Y)


def RelativePosition(X, Y=None):
    """The :grammar:`relative position of <vector> [from <vector>]` operator.

    If the :grammar:`from <vector>` is omitted, the position of ego is used.
    """
    X = toVector(X, '"relative position of X from Y" with X not a vector')
    if Y is None:
        Y = ego()
    Y = toVector(Y, '"relative position of X from Y" with Y not a vector')
    return X - Y


def RelativeHeading(X, Y=None):
    """The :grammar:`relative heading of <heading> [from <heading>]` operator.

    If the :grammar:`from <heading>` is omitted, the heading of ego is used.
    """
    X = toOrientation(
        X, '"relative heading of X from Y" with X not a heading or orientation'
    )
    if Y is None:
        Y = ego().orientation
    else:
        Y = toOrientation(Y, '"relative heading of X from Y" with Y not a heading')
    return normalizeAngle(X.yaw - Y.yaw)


def ApparentHeading(X, Y=None):
    """The :grammar:`apparent heading of <oriented point> [from <vector>]` operator.

    If the :grammar:`from <vector>` is omitted, the position of ego is used.
    """
    if not isA(X, OrientedPoint):
        raise TypeError('"apparent heading of X from Y" with X not an OrientedPoint')
    if Y is None:
        Y = ego()
    Y = toVector(Y, '"relative heading of X from Y" with Y not a vector')
    return apparentHeadingAtPoint(X.position, X.heading, Y)


def DistanceFrom(X, Y=None):
    """The :scenic:`distance from {X} to {Y}` polymorphic operator.

    Allowed forms::

        distance from <vector> [to <vector>]
        distance from <region> [to <vector>]
        distance from <vector> to <region>

    If the :grammar:`to <vector>` is omitted, the position of ego is used.
    """
    X = toTypes(
        X, (Vector, Region), '"distance from X to Y" with X neither a vector nor region'
    )
    if Y is None:
        Y = ego()
    Y = toTypes(
        Y, (Vector, Region), '"distance from X to Y" with Y neither a vector nor region'
    )
    return X.distanceTo(Y)


def DistancePast(X, Y=None):
    """The :grammar:`distance past <vector> of <oriented point>` operator.

    If the :grammar:`of {oriented point}` is omitted, the ego object is used.
    """
    X = toVector(X, '"distance past X" with X not a vector')
    if Y is None:
        Y = ego()
    Y = toType(Y, OrientedPoint, '"distance past X of Y" with Y not an OrientedPoint')
    return Y.distancePast(X)


# TODO(shun): Migrate to `AngleFrom`
def AngleTo(X):
    """The :grammar:`angle to <vector>` operator (using the position of ego as the reference)."""
    X = toVector(X, '"angle to X" with X not a vector')
    return ego().angleTo(X)


def AngleFrom(X=None, Y=None):
    """The :grammar:`angle from <vector> to <vector>` operator."""
    assert X is not None or Y is not None
    if X is None:
        X = ego()
    X = toVector(X, '"angle from X to Y" with X not a vector')
    if Y is None:
        Y = ego()
    Y = toVector(Y, '"angle from X to Y" with Y not a vector')
    return X.angleTo(Y)


def AltitudeTo(X):
    """The :grammar:`angle to <vector>` operator (using the position of ego as the reference)."""
    X = toVector(X, '"altitude to X" with X not a vector')
    return ego().altitudeTo(X)


def AltitudeFrom(X=None, Y=None):
    """The :grammar:`altitude from <vector> to <vector>` operator."""
    assert X is not None or Y is not None
    if X is None:
        X = ego()
    X = toVector(X, '"altitude from X to Y" with X not a vector')
    if Y is None:
        Y = ego()
    Y = toVector(Y, '"altitude from X to Y" with Y not a vector')
    return X.altitudeTo(Y)


def Follow(F, X, D):
    """The :grammar:`follow <field> from <vector> for <number>` operator."""
    if not isA(F, VectorField):
        raise TypeError('"follow F from X for D" with F not a vector field')
    X = toVector(X, '"follow F from X for D" with X not a vector')
    D = toScalar(D, '"follow F from X for D" with D not a number')
    pos = F.followFrom(X, D)
    orientation = F[pos]
    return OrientedPoint._with(position=pos, parentOrientation=orientation)


def VisibleFromOp(region, base):
    """The :grammar:`<region> visible from <point>` operator."""
    region = toType(region, Region, '"X visible from Y" with X not a Region')
    if not isA(base, Point):
        raise TypeError('"X visible from Y" with Y not a Point')
    return region.intersect(base.visibleRegion)


def NotVisibleFromOp(region, base):
    """The :grammar:`<region> not visible from <point>` operator."""
    region = toType(region, Region, '"X visible from Y" with X not a Region')
    if not isA(base, Point):
        raise TypeError('"X not visible from Y" with Y not a Point')

    return region.difference(base.visibleRegion)


def CanSee(X, Y):
    """The :scenic:`{X} can see {Y}` polymorphic operator.

    Allowed forms::

        <point> can see <vector>
        <point> can see <point>
    """
    if isActive():
        raise InvalidScenarioError(
            '"can see" operator prohibited at top level of Scenic programs'
        )

    if not isA(X, Point):
        raise TypeError('"X can see Y" with X not a Point, OrientedPoint, or Object')

    if not canCoerce(Y, Vector):
        raise TypeError('"X can see Y" with Y not a Vector, Point, or Object')

    objects = toDistribution(currentScenario._objects)

    @distributionFunction
    def canSeeHelper(X, Y, objects):
        if not isA(Y, Point):
            Y = toVector(
                Y, '"X can see Y" with X not a Vector, Point, OrientedPoint, or Object'
            )

        occludingObjects = tuple(
            obj for obj in objects if obj.occluding and X is not obj and Y is not obj
        )

        return X.canSee(Y, occludingObjects=occludingObjects)

    return canSeeHelper(X, Y, objects)


@distributionFunction
def Intersects(X, Y):
    """The :scenic:`{X} intersects {Y}` operator."""
    if isA(X, Object):
        return X.intersects(Y)
    else:
        return Y.intersects(X)


### Specifiers


def With(prop, val):
    """The :grammar:`with <property> <value>` specifier.

    Specifies the given property, with no dependencies.
    """
    return Specifier(f"With({prop})", {prop: 1}, {prop: val})


def At(pos):
    """The :grammar:`at <vector>` specifier.

    Specifies :prop:`position`, with no dependencies.
    """
    pos = toVector(pos, 'specifier "at X" with X not a vector')
    return Specifier("At", {"position": 1}, {"position": pos})


def In(region):
    """The :grammar:`in <region>` specifier.

    Specifies :prop:`position`, and optionally, :prop:`parentOrientation` if the given region
    has a preferred orientation, with no dependencies.
    """
    region = toType(region, Region, 'specifier "in R" with R not a Region')
    pos = Region.uniformPointIn(region)
    props = {"position": 1}
    values = {"position": pos}
    if alwaysProvidesOrientation(region):
        props["parentOrientation"] = 3
        values["parentOrientation"] = region.orientation[pos]
    return Specifier("In", props, values)


def ContainedIn(region):
    """The :grammar:`contained in <region>` specifier.

    Specifies :prop:`position`, :prop:`regionContainedIn`, and optionally, :prop:`parentOrientation`
    if the given region has a preferred orientation, with no dependencies.
    """
    region = toType(region, Region, 'specifier "contained in R" with R not a Region')
    pos = Region.uniformPointIn(region)
    props = {"position": 1, "regionContainedIn": 1}
    values = {"position": pos, "regionContainedIn": region}
    if alwaysProvidesOrientation(region):
        props["parentOrientation"] = 3
        values["parentOrientation"] = region.orientation[pos]
    return Specifier("ContainedIn", props, values)


def On(thing):
    """The :specifier:`on {X}` specifier.

    Specifies :prop:`position`, and optionally, :prop:`parentOrientation` if the given region
    has a preferred orientation. Depends on :prop:`onDirection`, :prop:`baseOffset`,
    and :prop:`contactTolerance`.

    Note that while :specifier:`on` can be used with `Region`, `Object` and `Vector`,
    it cannot be used with a distribution containing anything other than `Region`.

    May be used to modify an already-specified :prop:`position` property.

    Allowed forms:
        on <region>
        on <object>
        on <vector>
    """
    if isA(thing, Object):
        # Target is an Object: use its onSurface.
        target = thing.onSurface
    elif canCoerce(thing, Vector, exact=True):
        # Target is a vector
        target = toVector(thing)
    elif canCoerce(thing, Region):
        # Target is a region (or could theoretically be coerced to one),
        # so we can use it as a target.
        target = toType(thing, Region)
    else:
        raise TypeError('specifier "on R" with R not a Region, Object, or Vector')

    props = {"position": 1}

    if isA(target, Region) and alwaysProvidesOrientation(target):
        props["parentOrientation"] = 2

    def helper(context):
        # Pick position based on whether we are specifying or modifying
        if hasattr(context, "position"):
            if isA(target, Vector):
                raise TypeError('Cannot use modifying "on V" with V a vector.')

            pos = projectVectorHelper(target, context.position, context.onDirection)
        elif isA(target, Vector):
            pos = target
        else:
            pos = Region.uniformPointIn(target)

        values = {}

        contactOffset = Vector(0, 0, context.contactTolerance / 2) - context.baseOffset

        if "parentOrientation" in props:
            values["parentOrientation"] = target.orientation[pos]
            contactOffset = contactOffset.rotatedBy(values["parentOrientation"])

        values["position"] = pos + contactOffset

        return values

    return ModifyingSpecifier(
        "On",
        props,
        DelayedArgument({"onDirection", "baseOffset", "contactTolerance"}, helper),
        modifiable_props={"position"},
    )


@distributionFunction
def projectVectorHelper(region, pos, onDirection):
    on_pos = region.projectVector(pos, onDirection=onDirection)

    if on_pos is None:
        raise RejectionException("Unable to place object on surface.")
    else:
        return on_pos


def alwaysProvidesOrientation(region):
    """Whether a Region or distribution over Regions always provides an orientation."""
    if isinstance(region, Region):
        return region.orientation is not None
    elif isinstance(region, MultiplexerDistribution) and all(
        alwaysProvidesOrientation(opt) for opt in region.options
    ):
        return True
    else:  # TODO improve somehow!
        try:
            sample = region.sample()
            return sample.orientation is not None or sample is nowhere
        except RejectionException:
            return False


def OffsetBy(offset):
    """The :grammar:`offset by <vector>` specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, with no dependencies.
    """
    offset = toVector(offset, 'specifier "offset by X" with X not a vector')
    value = {
        "position": RelativeTo(offset, ego()).toVector(),
        "parentOrientation": ego().orientation,
    }
    return Specifier("OffsetBy", {"position": 1, "parentOrientation": 3}, value)


def OffsetAlongSpec(direction, offset):
    """The :specifier:`offset along {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, with no dependencies.

    Allowed forms::

        offset along <heading> by <vector>
        offset along <field> by <vector>
    """
    pos = OffsetAlong(ego(), direction, offset)
    parentOrientation = ego().orientation
    return Specifier(
        "OffsetAlong",
        {"position": 1, "parentOrientation": 3},
        {"position": pos, "parentOrientation": parentOrientation},
    )


def Beyond(pos, offset, fromPt=None):
    """The :specifier:`beyond {X} by {Y} from {Z}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, with no dependencies.

    Allowed forms::

        beyond <vector> by <number> [from <vector>]
        beyond <vector> by <vector> [from <vector>]

    If the :grammar:`from <vector>` is omitted, the position of ego is used.
    """
    # Ensure X can be coerced into vector form
    pos = toVector(pos, 'specifier "beyond X by Y" with X not a vector')

    # If no from vector is specified, assume ego
    if fromPt is None:
        fromPt = ego()

    fromPt = toVector(fromPt, 'specifier "beyond X by Y from Z" with Z not a vector')

    dType = underlyingType(offset)

    if dType is builtins.float or dType is builtins.int:
        offset = Vector(0, offset, 0)
    else:
        # offset is not float or int, so try to coerce it into vector form.
        offset = toVector(
            offset, 'specifier "beyond X by Y" with X not a number or vector'
        )

    # If the from vector is oriented, set that to orientation. Else assume global coords.
    if isA(fromPt, OrientedPoint):
        orientation = fromPt.orientation
    else:
        orientation = Orientation.fromEuler(0, 0, 0)

    direction = pos - fromPt
    sphericalCoords = direction.sphericalCoordinates()
    offsetRotation = Orientation.fromEuler(sphericalCoords[1], sphericalCoords[2], 0)

    new_direction = pos + offset.applyRotation(offsetRotation)

    return Specifier(
        "Beyond",
        {"position": 1, "parentOrientation": 3},
        {"position": new_direction, "parentOrientation": orientation},
    )


def VisibleFrom(base):
    """The :grammar:`visible from <point>` specifier.

    Specifies :prop:`_observingEntity` and :prop:`position`, with no dependencies.
    """
    if not isA(base, Point):
        raise TypeError('specifier "visible from O" with O not a Point')

    def helper(self):
        if mode2D:
            position = Region.uniformPointIn(base.visibleRegion)
        else:
            containing_region = (
                currentScenario._workspace.region
                if self.regionContainedIn is None
                and currentScenario._workspace is not None
                else self.regionContainedIn
            )
            position = (
                Region.uniformPointIn(everywhere, tag="visible")
                if containing_region is None
                else Region.uniformPointIn(containing_region)
            )

        return {"position": position, "_observingEntity": base}

    return Specifier(
        "Visible/VisibleFrom",
        {"position": 3, "_observingEntity": 1},
        DelayedArgument({"regionContainedIn"}, helper),
    )


def VisibleSpec():
    """The :specifier:`visible` specifier (equivalent to :specifier:`visible from ego`).

    Specifies :prop:`_observingEntity` and :prop:`position`, with no dependencies.
    """
    return VisibleFrom(ego())


def NotVisibleFrom(base):
    """The :grammar:`not visible from <point>` specifier.

    Specifies :prop:`_nonObservingEntity` and :prop:`position`, depending on :prop:`regionContainedIn`.

    See `VisibleFrom`.
    """
    if not isA(base, Point):
        raise TypeError('specifier "not visible from O" with O not a Point')

    def helper(self):
        region = self.regionContainedIn
        if region is None:
            if currentScenario._workspace is None:
                raise InvalidScenarioError(
                    '"not visible" specifier with no workspace or containing region defined'
                )
            region = currentScenario._workspace.region

        if mode2D:
            position = Region.uniformPointIn(region.difference(base.visibleRegion))
        else:
            # We can't limit the available region since any spot could potentially be occluded.
            position = Region.uniformPointIn(convertToFootprint(region))

        return {"position": position, "_nonObservingEntity": base}

    return Specifier(
        "NotVisible/NotVisibleFrom",
        {"position": 3, "_nonObservingEntity": 1},
        DelayedArgument({"regionContainedIn"}, helper),
    )


def NotVisibleSpec():
    """The :specifier:`not visible` specifier (equivalent to :specifier:`not visible from ego`).

    Specifies :prop:`_nonObservingEntity` and :prop:`position`, depending on :prop:`regionContainedIn`.
    """
    return NotVisibleFrom(ego())


def LeftSpec(pos, dist=None):
    """The :specifier:`left of {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally, :prop:`parentOrientation`, depending on :prop:`width`.

    Allowed forms::

        left of <oriented point> [by <scalar/vector>]
        left of <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, the object's contact tolerance is used.
    """
    return directionalSpecHelper(
        "Left of",
        pos,
        dist,
        "width",
        lambda dist: (dist, 0, 0),
        lambda self, dims, tol, dx, dy, dz: Vector(
            -self.width / 2 - dx - dims[0] / 2 - tol, dy, dz
        ),
    )


def RightSpec(pos, dist=None):
    """The :specifier:`right of {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, depending on :prop:`width`.

    Allowed forms::

        right of <oriented point> [by <scalar/vector>]
        right of <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, zero is used.
    """
    return directionalSpecHelper(
        "Right of",
        pos,
        dist,
        "width",
        lambda dist: (dist, 0, 0),
        lambda self, dims, tol, dx, dy, dz: Vector(
            self.width / 2 + dx + dims[0] / 2 + tol, dy, dz
        ),
    )


def Ahead(pos, dist=None):
    """The :specifier:`ahead of {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, depending on :prop:`length`.

    Allowed forms::

        ahead of <oriented point> [by <scalar/vector>]
        ahead of <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, the object's contact tolerance is used.
    """
    return directionalSpecHelper(
        "Ahead of",
        pos,
        dist,
        "length",
        lambda dist: (0, dist, 0),
        lambda self, dims, tol, dx, dy, dz: Vector(
            dx, self.length / 2 + dy + dims[1] / 2 + tol, dz
        ),
    )


def Behind(pos, dist=None):
    """The :specifier:`behind {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, depending on :prop:`length`.

    Allowed forms::

        behind <oriented point> [by <scalar/vector>]
        behind <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, the object's contact tolerance is used.
    """
    return directionalSpecHelper(
        "Behind",
        pos,
        dist,
        "length",
        lambda dist: (0, dist, 0),
        lambda self, dims, tol, dx, dy, dz: Vector(
            dx, -self.length / 2 - dy - dims[1] / 2 - tol, dz
        ),
    )


def Above(pos, dist=None):
    """The :specifier:`above {X} by {Y}` polymorphic specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, depending on :prop:`height`.

    Allowed forms::

        above <oriented point> [by <scalar/vector>]
        above <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, the object's contact tolerance is used.
    """
    return directionalSpecHelper(
        "Above",
        pos,
        dist,
        "height",
        lambda dist: (0, 0, dist),
        lambda self, dims, tol, dx, dy, dz: Vector(
            dx, dy, self.height / 2 + dz + dims[2] / 2 + tol
        ),
    )


def Below(pos, dist=None):
    """The :specifier:`below {X} by {Y}` polymorphic specifier.

    Specifies :prop`position`, and optionally :prop:`parentOrientation`, depending on :prop:`height`.

    Allowed forms::

        below <oriented point> [by <scalar/vector>]
        below <vector> [by <scalar/vector>]

    If the :grammar:`by <scalar/vector>` is omitted, the object's contact tolerance is used.
    """
    return directionalSpecHelper(
        "Below",
        pos,
        dist,
        "height",
        lambda dist: (0, 0, dist),
        lambda self, dims, tol, dx, dy, dz: Vector(
            dx, dy, -self.height / 2 - dz - dims[2] / 2 - tol
        ),
    )


def directionalSpecHelper(syntax, pos, dist, axis, toComponents, makeOffset):
    prop = {"position": 1}
    if dist is None:
        dx = dy = dz = 0
    elif canCoerce(dist, builtins.float):
        dx, dy, dz = toComponents(coerce(dist, builtins.float))
    elif canCoerce(dist, Vector):
        dx, dy, dz = coerce(dist, Vector)
    else:
        raise TypeError(f'"{syntax} X by D" with D not a number or vector')

    @distributionFunction
    def makeContactOffset(dist, ct):
        if dist is None:
            return ct / 2
        else:
            return 0

    if isA(pos, Object):
        prop["parentOrientation"] = 3
        obj_dims = (pos.width, pos.length, pos.height)
        val = lambda self: {
            "position": pos.relativePosition(
                makeOffset(
                    self,
                    obj_dims,
                    makeContactOffset(dist, self.contactTolerance),
                    dx,
                    dy,
                    dz,
                )
            ),
            "parentOrientation": pos.orientation,
        }
        new = DelayedArgument({axis, "contactTolerance"}, val)
    elif isA(pos, OrientedPoint):
        prop["parentOrientation"] = 3
        val = lambda self: {
            "position": pos.relativePosition(makeOffset(self, (0, 0, 0), 0, dx, dy, dz)),
            "parentOrientation": pos.orientation,
        }
        new = DelayedArgument({axis}, val)
    else:
        pos = toVector(pos, f'specifier "{syntax} X" with X not a vector')
        val = lambda self: {
            "position": pos.offsetLocally(
                self.orientation, makeOffset(self, (0, 0, 0), 0, dx, dy, dz)
            )
        }
        new = DelayedArgument({axis, "orientation"}, val)
    return Specifier(syntax, prop, new)


def Following(field, dist, fromPt=None):
    """The :specifier:`following {F} from {X} for {D}` specifier.

    Specifies :prop:`position`, and optionally :prop:`parentOrientation`, with no dependencies.

    Allowed forms::

        following <field> [from <vector>] for <number>

    If the :grammar:`from <vector>` is omitted, the position of ego is used.
    """
    if fromPt is None:
        fromPt = ego()
    field = toType(field, VectorField)
    fromPt = toVector(fromPt, '"following F from X for D" with X not a vector')
    dist = toScalar(dist, '"following F for D" with D not a number')
    pos = field.followFrom(fromPt, dist)
    orientation = field[pos]
    return Specifier(
        "Following",
        {"position": 1, "parentOrientation": 3},
        {"position": pos, "parentOrientation": orientation},
    )


def Facing(heading):
    """The :specifier:`facing {X}` polymorphic specifier.

    Specifies :prop:`yaw`, :prop:`pitch`, and :prop:`roll`, depending on :prop:`parentOrientation`,
    and depending on the form::

        facing <number>     # no further dependencies;
        facing <field>      # depends on 'position'
    """
    if isA(heading, VectorField):

        def helper(context):
            headingAtPos = heading[context.position]
            if alwaysGlobalOrientation(context.parentOrientation):
                orientation = headingAtPos  # simplify expr tree in common case
            else:
                orientation = context.parentOrientation.inverse * headingAtPos
            return {
                "yaw": orientation.yaw,
                "pitch": orientation.pitch,
                "roll": orientation.roll,
            }

        return Specifier(
            "Facing",
            {"yaw": 1, "pitch": 1, "roll": 1},
            DelayedArgument({"position", "parentOrientation"}, helper),
        )
    else:
        orientation = toOrientation(
            heading, "facing x with x not a heading or orientation"
        )
        orientationDeps = requiredProperties(orientation)

        def helper(context):
            target_orientation = valueInContext(orientation, context)
            euler = context.parentOrientation.localAnglesFor(target_orientation)
            return {"yaw": euler[0], "pitch": euler[1], "roll": euler[2]}

        return Specifier(
            "Facing",
            {"yaw": 1, "pitch": 1, "roll": 1},
            DelayedArgument({"parentOrientation"} | orientationDeps, helper),
        )


def FacingToward(pos):
    """The :grammar:`facing toward <vector>` specifier.

    Specifies :prop:`yaw`, depending on :prop:`position` and :prop:`parentOrientation`.
    """
    pos = toVector(pos, 'specifier "facing toward X" with X not a vector')

    def helper(context):
        direction = pos - context.position
        rotated = direction.applyRotation(context.parentOrientation.inverse)
        sphericalCoords = (
            rotated.sphericalCoordinates()
        )  # Ignore the rho, sphericalCoords[0]
        return {"yaw": sphericalCoords[1]}

    return Specifier(
        "FacingToward",
        {"yaw": 1},
        DelayedArgument({"position", "parentOrientation"}, helper),
    )


def FacingDirectlyToward(pos):
    """The :grammar:`facing directly toward <vector>` specifier.

    Specifies :prop:`yaw` and :prop:`pitch`, depends on :prop:`position` and :prop:`parentOrientation`.
    """
    pos = toVector(pos, 'specifier "facing directly toward X" with X not a vector')

    def helper(context):
        """
        Same process as above, except by default also specify the pitch euler angle
        """
        direction = pos - context.position
        rotated = direction.applyRotation(context.parentOrientation.inverse)
        sphericalCoords = rotated.sphericalCoordinates()
        return {"yaw": sphericalCoords[1], "pitch": sphericalCoords[2]}

    return Specifier(
        "FacingDirectlyToward",
        {"yaw": 1, "pitch": 1},
        DelayedArgument({"position", "parentOrientation"}, helper),
    )


def FacingAwayFrom(pos):
    """The :grammar:`facing away from <vector>` specifier.

    Specifies :prop:`yaw`, depending on :prop:`position` and :prop:`parentOrientation`.
    """
    pos = toVector(pos, 'specifier "facing away from X" with X not a vector')

    def helper(context):
        """
        As in FacingToward, except invert the resulting rotation axis
        """
        direction = context.position - pos
        rotated = direction.applyRotation(context.parentOrientation.inverse)
        sphericalCoords = rotated.sphericalCoordinates()
        return {"yaw": sphericalCoords[1]}

    return Specifier(
        "FacingAwayFrom",
        {"yaw": 1},
        DelayedArgument({"position", "parentOrientation"}, helper),
    )


def FacingDirectlyAwayFrom(pos):
    """The :grammar:`facing directly away from <vector>` specifier.

    Specifies :prop:`yaw` and :prop:`pitch`, depending on :prop:`position` and :prop:`parentOrientation`.
    """
    pos = toVector(pos, 'specifier "facing away from X" with X not a vector')

    def helper(context):
        direction = context.position - pos
        rotated = direction.applyRotation(context.parentOrientation.inverse)
        sphericalCoords = rotated.sphericalCoordinates()
        return {"yaw": sphericalCoords[1], "pitch": sphericalCoords[2]}

    return Specifier(
        "FacingDirectlyToward",
        {"yaw": 1, "pitch": 1},
        DelayedArgument({"position", "parentOrientation"}, helper),
    )


def ApparentlyFacing(heading, fromPt=None):
    """The :grammar:`apparently facing <heading> [from <vector>]` specifier.

    Specifies :prop:`yaw`, depending on :prop:`position` and :prop:`parentOrientation`.

    If the :grammar:`from <vector>` is omitted, the position of ego is used.
    """
    heading = toHeading(heading, 'specifier "apparently facing X" with X not a heading')
    if fromPt is None:
        fromPt = ego()
    fromPt = toVector(
        fromPt, 'specifier "apparently facing X from Y" with Y not a vector'
    )

    def helper(context):
        return {"yaw": fromPt.angleTo(context.position) + heading}

    return Specifier(
        "ApparentlyFacing",
        {"yaw": 1},
        DelayedArgument({"position", "parentOrientation"}, helper),
    )


### Primitive functions overriding Python builtins

# N.B. applying functools.wraps to preserve the metadata of the original
# functions seems to break pickling/unpickling


@distributionFunction
def filter(function, iterable):
    return list(builtins.filter(function, iterable))


@distributionFunction
def str(*args, **kwargs):
    return builtins.str(*args, **kwargs)


@distributionFunction
def float(*args, **kwargs):
    return builtins.float(*args, **kwargs)


@distributionFunction
def int(*args, **kwargs):
    return builtins.int(*args, **kwargs)


@distributionFunction
def round(*args, **kwargs):
    return builtins.round(*args, **kwargs)


def len(obj):
    return obj.__len__()


def range(*args):
    if any(needsSampling(arg) for arg in args):
        raise RandomControlFlowError("cannot construct a range with random parameters")
    return builtins.range(*args)


### Temporal Operators Factories


def AtomicProposition(closure, syntaxId):
    return propositions.Atomic(closure, syntaxId)


def PropositionAnd(reqs):
    return propositions.And(reqs)


def PropositionOr(reqs):
    return propositions.Or(reqs)


def PropositionNot(req):
    return propositions.Not(req)


def Always(req):
    return propositions.Always(req)


def Eventually(req):
    return propositions.Eventually(req)


def Next(req):
    return propositions.Next(req)


def Until(lhs, rhs):
    return propositions.Until(lhs, rhs)


def Implies(lhs, rhs):
    return propositions.Implies(lhs, rhs)



"""Interface to the LGSVL driving simulator.

The LGSVL Simulator interface was deprecated in Scenic 3. To continue to use the interface, please use Scenic 2.
"""

# Only import LGSVLSimulator if the lgsvl package is installed; otherwise the
# import would raise an exception.
lgsvl = None
try:
    import lgsvl
except ImportError:
    pass
if lgsvl:
    from .simulator import LGSVLSimulator
del lgsvl



"""A basic color type.

This used for example to represent car colors in the abstract driving domain,
as well as in the interfaces to GTA and Webots.
"""

from collections import namedtuple
import colorsys
import random
import struct

from scenic.core.distributions import Distribution, Normal, Options, Range, toDistribution
from scenic.core.lazy_eval import valueInContext
from scenic.core.object_types import Mutator


class Color(namedtuple("Color", ["r", "g", "b"])):
    """A color as an RGB tuple."""

    @classmethod
    def withBytes(cls, color):
        return cls._make(c / 255.0 for c in color)

    @staticmethod
    def realToByte(color):
        return tuple(int(round(255 * c)) for c in color)

    @staticmethod
    def encodeTo(color, stream):
        stream.write(struct.pack("<ddd", *color))

    @staticmethod
    def decodeFrom(stream):
        return Color(*struct.unpack("<ddd", stream.read(24)))

    @staticmethod
    def uniformColor():
        """Return a uniformly random color."""
        return toDistribution(Color(Range(0, 1), Range(0, 1), Range(0, 1)))

    @staticmethod
    def defaultCarColor():
        """Default color distribution for cars.

        The distribution starts with a base distribution over 9 discrete colors,
        then adds Gaussian HSL noise. The base distribution uses color popularity
        statistics from a `2012 DuPont survey`_.

        .. _2012 DuPont survey: https://web.archive.org/web/20121229065631/http://www2.dupont.com/Media_Center/en_US/color_popularity/Images_2012/DuPont2012ColorPopularity.pdf
        """
        baseColors = {
            (248, 248, 248): 0.24,  # white
            (50, 50, 50): 0.19,  # black
            (188, 185, 183): 0.16,  # silver
            (130, 130, 130): 0.15,  # gray
            (194, 92, 85): 0.10,  # red
            (75, 119, 157): 0.07,  # blue
            (197, 166, 134): 0.05,  # brown/beige
            (219, 191, 105): 0.02,  # yellow/gold
            (68, 160, 135): 0.02,  # green
        }
        converted = {Color.withBytes(color): prob for color, prob in baseColors.items()}
        baseColor = Options(converted)
        # TODO improve this?
        hueNoise = Normal(0, 0.1)
        satNoise = Normal(0, 0.1)
        lightNoise = Normal(0, 0.1)
        return NoisyColorDistribution(baseColor, hueNoise, satNoise, lightNoise)


class NoisyColorDistribution(Distribution):
    """A distribution given by HSL noise around a base color.

    Arguments:
        baseColor (RGB tuple): base color
        hueNoise (float): noise to add to base hue
        satNoise (float): noise to add to base saturation
        lightNoise (float): noise to add to base lightness
    """

    def __init__(self, baseColor, hueNoise, satNoise, lightNoise):
        super().__init__(baseColor, hueNoise, satNoise, lightNoise, valueType=Color)
        self.baseColor = baseColor
        self.hueNoise = hueNoise
        self.satNoise = satNoise
        self.lightNoise = lightNoise

    @staticmethod
    def addNoiseTo(color, hueNoise, lightNoise, satNoise):
        try:
            hue, lightness, saturation = colorsys.rgb_to_hls(*color[:3])
        except ZeroDivisionError:
            hue, lightness, saturation = 0.0, 1.0, 0.0
        hue = max(0, min(1, hue + hueNoise))
        lightness = max(0, min(1, lightness + lightNoise))
        saturation = max(0, min(1, saturation + satNoise))
        return colorsys.hls_to_rgb(hue, lightness, saturation)

    def sampleGiven(self, value):
        bc = value[self.baseColor]
        return Color(
            *self.addNoiseTo(
                bc, value[self.hueNoise], value[self.lightNoise], value[self.satNoise]
            )
        )

    def evaluateInner(self, context):
        self.baseColor = valueInContext(self.baseColor, context)
        self.hueNoise = valueInContext(self.hueNoise, context)
        self.satNoise = valueInContext(self.satNoise, context)
        self.lightNoise = valueInContext(self.lightNoise, context)


class ColorMutator(Mutator):
    """Mutator that adds Gaussian HSL noise to the ``color`` property."""

    def appliedTo(self, obj):
        stddev = 0.05 * obj.mutationScale
        hueNoise = random.gauss(0, stddev)
        satNoise = random.gauss(0, stddev)
        lightNoise = random.gauss(0, stddev)
        obj.color = NoisyColorDistribution.addNoiseTo(
            obj.color, hueNoise, lightNoise, satNoise
        )
        return (obj, True)  # allow further mutation



"""Python library supporting the main Scenic module."""

import itertools
import math
import sys
import time

import numpy as np
import shapely.geometry
import shapely.ops

from scenic.core.geometry import (
    cleanChain,
    normalizeAngle,
    plotPolygon,
    polygonUnion,
    rotateVector,
)
from scenic.core.regions import PolygonalRegion, PolylineRegion, nowhere
from scenic.core.vectors import PolygonalVectorField
from scenic.core.workspaces import Workspace
import scenic.simulators.webots.world_parser as world_parser
from scenic.syntax.veneer import verbosePrint


def polygonWithPoints(points):
    polygon = shapely.geometry.Polygon(points)
    if not polygon.is_valid:  # TODO improve hack?
        verbosePrint(f"WARNING: simplifying invalid polygon with points {points}")
        polygon = polygon.simplify(0.5)
        if not polygon.is_valid:
            raise RuntimeError(f"unable to simplify polygon {polygon}")
    return polygon


def regionWithPolygons(polygons, orientation=None):
    if polygons.is_empty:
        return nowhere
    else:
        return PolygonalRegion(polygon=polygons, orientation=orientation)


## Classes for WBT nodes we are interested in


class OSMObject:
    """Objects with OSM id tags"""

    def __init__(self, attrs):
        self.attrs = attrs
        self.osmID = attrs["id"]


class Road(OSMObject):
    """OSM roads"""

    def __init__(self, attrs, driveOnLeft=False):
        super().__init__(attrs)
        self.driveOnLeft = driveOnLeft
        self.translation = attrs["translation"]
        pts = [
            np.array(webotsToScenicPosition(p + self.translation))
            for p in attrs["wayPoints"]
        ]
        self.waypoints = tuple(cleanChain(pts, 0.05))
        assert len(self.waypoints) > 1, pts
        self.width = float(attrs.get("width", 7))
        self.lanes = int(attrs.get("numberOfLanes", 2))
        if self.lanes < 1:
            raise RuntimeError(f"Road {self.osmID} has fewer than 1 lane!")
        self.forwardLanes = int(attrs.get("numberOfForwardLanes", 1))
        # if self.forwardLanes < 1:
        #   raise RuntimeError(f'Road {self.osmID} has fewer than 1 forward lane!')
        self.backwardLanes = self.lanes - self.forwardLanes
        self.hasLeftSidewalk = attrs.get("leftBorder", True)
        self.hasRightSidewalk = attrs.get("rightBorder", True)
        self.sidewalkWidths = list(attrs.get("roadBorderWidth", [0.8]))
        if (self.hasLeftSidewalk or self.hasRightSidewalk) and len(
            self.sidewalkWidths
        ) < 1:
            raise RuntimeError(f"Road {self.osmID} has sidewalk with empty width!")
        self.startCrossroad = attrs.get("startJunction")
        self.endCrossroad = attrs.get("endJunction")

    def computeGeometry(self, crossroads, snapTolerance=0.05):
        ## Approximate bounding polygon and sidewalks
        # TODO improve this!!!
        lefts, rights = [], []
        leftSidewalk, rightSidewalk = [], []
        headings = []
        sidewalkWidths = itertools.chain(
            self.sidewalkWidths, itertools.repeat(self.sidewalkWidths[-1])
        )
        segments = zip(self.waypoints, sidewalkWidths)
        for i, segment in enumerate(segments):
            point, sidewalkWidth = segment
            if i + 1 < len(self.waypoints):
                nextPt = self.waypoints[i + 1]
                dx, dy = nextPt - point
                heading = normalizeAngle(math.atan2(dy, dx) - (math.pi / 2))
                headings.append(heading)
                perp = np.array([-dy, dx])
                perp /= np.linalg.norm(perp)
            else:
                pass  # use perp from last segment
            toEdge = perp * (self.width / 2)
            left = point + toEdge
            right = point - toEdge
            lefts.append(left)
            rights.append(right)
            toEdge = perp * sidewalkWidth
            leftSidewalk.append(left + toEdge)
            rightSidewalk.append(right - toEdge)

        # Snap to adjacent crossroads if possible
        if snapTolerance > 0:
            sc = self.startCrossroad
            if sc is not None:
                if sc not in crossroads:
                    raise RuntimeError(
                        f"Road {self.osmID} begins at invalid crossroad {sc}"
                    )
                crossroad = crossroads[sc]
                if crossroad.region is not None:
                    pt = shapely.geometry.Point(lefts[0])
                    pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
                    lefts[0] = np.array([pt.x, pt.y])
                    pt = shapely.geometry.Point(rights[0])
                    pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
                    rights[0] = np.array([pt.x, pt.y])
                    perp = lefts[0] - rights[0]
                    toEdge = perp * (self.sidewalkWidths[0] / np.linalg.norm(perp))
                    leftSidewalk[0] = lefts[0] + toEdge
                    rightSidewalk[0] = rights[0] - toEdge
            ec = self.endCrossroad
            if ec is not None:
                if ec not in crossroads:
                    raise RuntimeError(
                        f"Road {self.osmID} ends at invalid crossroad {ec}"
                    )
                crossroad = crossroads[ec]
                if crossroad.region is not None:
                    pt = shapely.geometry.Point(lefts[-1])
                    pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
                    lefts[-1] = np.array([pt.x, pt.y])
                    pt = shapely.geometry.Point(rights[-1])
                    pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
                    rights[-1] = np.array([pt.x, pt.y])
                    perp = lefts[-1] - rights[-1]
                    toEdge = perp * (self.sidewalkWidths[-1] / np.linalg.norm(perp))
                    leftSidewalk[-1] = lefts[-1] + toEdge
                    rightSidewalk[-1] = rights[-1] - toEdge

        roadPoints = lefts + list(reversed(rights))
        self.leftCurb = PolylineRegion(reversed(lefts))
        self.rightCurb = PolylineRegion(rights)

        self.leftSidewalk = self.rightSidewalk = None
        if self.hasLeftSidewalk:
            points = lefts + list(reversed(leftSidewalk))
            polygon = polygonWithPoints(points)
            assert polygon.is_valid, self.waypoints
            self.leftSidewalk = PolygonalRegion(polygon=polygon)
        if self.hasRightSidewalk:
            points = rights + list(reversed(rightSidewalk))
            polygon = polygonWithPoints(points)
            assert polygon.is_valid, self.waypoints
            self.rightSidewalk = PolygonalRegion(polygon=polygon)

        ## Compute lanes and traffic directions
        cells = []
        la, ra = lefts[0], rights[0]
        gapA = (ra - la) / self.lanes
        markerA = ra
        laneMarkers = [[] for lane in range(self.lanes)]
        for lb, rb, heading in zip(lefts[1:], rights[1:], headings):
            # Compute lanes for this segment of road
            gapB = (rb - lb) / self.lanes
            markerB = rb
            for lane, markers in enumerate(laneMarkers):
                forward = lane < self.forwardLanes
                if self.driveOnLeft:
                    forward = not forward
                nextMarkerA = markerA - gapA
                nextMarkerB = markerB - gapB
                markers.append(nextMarkerA)
                cell = shapely.geometry.Polygon(
                    (markerA, markerB, nextMarkerB, nextMarkerA)
                )
                heading = heading if forward else normalizeAngle(heading + math.pi)
                cells.append((cell, heading))
                markerA = nextMarkerA
                markerB = nextMarkerB
            gapA = gapB
            markerA = rb
        self.lanes = []
        markerB = rb
        rightEdge = rights
        for lane, markers in enumerate(laneMarkers):
            markerB = markerB - gapB
            markers.append(markerB)
            self.lanes.append(PolygonalRegion(rightEdge + list(reversed(markers))))
            rightEdge = markers
        self.laneMarkers = laneMarkers[:-1]
        self.cells = cells
        self.direction = PolygonalVectorField(f"Road{self.osmID}Direction", cells)

        roadPolygon = polygonWithPoints(roadPoints)
        self.region = PolygonalRegion(polygon=roadPolygon, orientation=self.direction)

    def show(self, plt):
        if self.hasLeftSidewalk:
            x, y = zip(*self.leftSidewalk.points)
            plt.fill(x, y, "#A0A0FF")
        if self.hasRightSidewalk:
            x, y = zip(*self.rightSidewalk.points)
            plt.fill(x, y, "#A0A0FF")
        self.region.show(plt, style="r:")
        x, y = zip(*self.lanes[0].points)
        plt.fill(x, y, color=(0.8, 1.0, 0.8))
        for lane, markers in enumerate(self.laneMarkers):
            x, y = zip(*markers)
            color = (0.8, 0.8, 0) if lane == self.backwardLanes - 1 else (0.3, 0.3, 0.3)
            plt.plot(x, y, "--", color=color)


class Crossroad(OSMObject):
    """OSM crossroads"""

    def __init__(self, attrs):
        super().__init__(attrs)
        self.translation = attrs["translation"]
        points = list(
            np.array(webotsToScenicPosition(p + self.translation)) for p in attrs["shape"]
        )
        if len(points) > 0:
            self.points = points
            self.region = PolygonalRegion(self.points)
        else:
            verbosePrint(f"WARNING: Crossroad {self.osmID} has empty shape field!")
            self.region = None

    def show(self, plt):
        if self.region is not None:
            x, y = zip(*self.points)
            plt.fill(x, y, color=(1, 0.9, 0.9))
            plt.plot(x, y, ":", color=(1, 0.5, 0))


class PedestrianCrossing:
    """PedestrianCrossing nodes"""

    def __init__(self, attrs):
        self.translation = attrs.get("translation", np.array((0, 0, 0)))
        pos = np.array(webotsToScenicPosition(self.translation))
        name = attrs.get("name", "")
        self.angle = webotsToScenicRotation(attrs.get("rotation", (0, 1, 0, 0)))
        if self.angle is None:
            raise RuntimeError(f'PedestrianCrossing "{name}" is not 2D!')
        size = attrs.get("size", (20, 8))
        self.length, self.width = float(size[0]), float(size[1])
        self.length += 0.2  # pad length to intersect sidewalks better  # TODO improve?
        hl, hw = self.length / 2, self.width / 2
        self.corners = tuple(
            pos + rotateVector(vec, -self.angle)
            for vec in ((hl, hw), (-hl, hw), (-hl, -hw), (hl, -hw))
        )
        self.region = PolygonalRegion(self.corners)

    def show(self, plt):
        x, y = zip(*self.corners)
        plt.fill(x, y, color="#A0A0FF")


## Workspace derived from a WBT world


class WebotsWorkspace(Workspace):
    def __init__(self, world):
        # Find roads, crossroads, and pedestrian crossings
        nodeClasses = {
            "Road": Road,
            "Crossroad": Crossroad,
            "PedestrianCrossing": PedestrianCrossing,
        }
        self.roads, self.crossroads, self.crossings = world_parser.findNodeTypesIn(
            ("Road", "Crossroad", "PedestrianCrossing"), world, nodeClasses
        )

        # Compute road geometry
        crossroadsByID = {crossroad.osmID: crossroad for crossroad in self.crossroads}
        for road in self.roads:
            road.computeGeometry(crossroadsByID)

        # Construct regions
        allCells = []
        drivableAreas = []
        for road in self.roads:
            assert road.region.polygons.is_valid, (road.waypoints, road.region.points)
            allCells.extend(road.cells)
        for crossroad in self.crossroads:
            if crossroad.region is not None:
                for poly in crossroad.region.polygons.geoms:
                    allCells.append((poly, None))
        if not allCells:
            raise RuntimeError("Webots world has no drivable geometry!")
        self.roadDirection = PolygonalVectorField(
            "roadDirection",
            allCells,
            headingFunction=lambda pos: 0,
            defaultHeading=0,  # TODO fix
        )
        if not self.roads:
            roadPoly = None
            self.roadsRegion = nowhere
            self.curbsRegion = nowhere
        else:
            roadPoly = polygonUnion(road.region.polygons for road in self.roads)
            self.roadsRegion = PolygonalRegion(
                polygon=roadPoly, orientation=self.roadDirection
            )
            drivableAreas.append(roadPoly)
            curbs = [road.leftCurb.lineString for road in self.roads]
            curbs.extend(road.rightCurb.lineString for road in self.roads)
            curbLine = shapely.ops.unary_union(curbs)
            self.curbsRegion = PolylineRegion(polyline=curbLine)
        if not self.crossroads:
            crossroadPoly = None
            self.crossroadsRegion = nowhere
        else:
            crossroadPoly = polygonUnion(
                cr.region.polygons for cr in self.crossroads if cr.region is not None
            )
            self.crossroadsRegion = PolygonalRegion(
                polygon=crossroadPoly, orientation=self.roadDirection
            )
            drivableAreas.append(crossroadPoly)

        sidewalks = []
        walkableAreas = []
        for road in self.roads:
            if road.hasLeftSidewalk:
                sidewalks.append(road.leftSidewalk.polygons)
            if road.hasRightSidewalk:
                sidewalks.append(road.rightSidewalk.polygons)
        if not sidewalks:
            sidewalksPoly = None
            self.sidewalksRegion = nowhere
        else:
            sidewalksPoly = polygonUnion(sidewalks)
            self.sidewalksRegion = regionWithPolygons(sidewalksPoly)
            walkableAreas.append(sidewalksPoly)

        if not self.crossings:
            crossingsPoly = None
            self.crossingsRegion = nowhere
        else:
            crossingsPoly = polygonUnion(
                crossing.region.polygons for crossing in self.crossings
            )
            self.crossingsRegion = regionWithPolygons(crossingsPoly)
            walkableAreas.append(crossingsPoly)

        if not walkableAreas:
            self.walkableRegion = nowhere
        else:
            walkablePoly = polygonUnion(walkableAreas)
            self.walkableRegion = regionWithPolygons(walkablePoly)
        drivablePoly = polygonUnion(drivableAreas)
        self.drivableRegion = PolygonalRegion(
            polygon=drivablePoly, orientation=self.roadDirection
        )
        workspacePoly = polygonUnion(drivableAreas + walkableAreas)
        self.workspaceRegion = PolygonalRegion(polygon=workspacePoly)

        # Identify various roads and lanes of interest
        slowCurbs = []
        for road in self.roads:
            if road.forwardLanes > 1:
                slowCurbs.append(road.rightCurb)
            if road.backwardLanes > 1:
                slowCurbs.append(road.leftCurb)
        self.slowCurbs = slowCurbs

        super().__init__(self.workspaceRegion)

    def show2D(self, plt):
        self.drivableRegion.show(plt)
        for road in self.roads:
            road.show(plt)
        for crossroad in self.crossroads:
            crossroad.show(plt)
        for crossing in self.crossings:
            crossing.show(plt)
        for curb in self.slowCurbs:
            curb.show(plt, style="b-")

    @property
    def minimumZoomSize(self):
        return 30


# Old functions kept for backwards-compatibility


def webotsToScenicPosition(pos):
    """Convert a Webots position to a Scenic position.
    Drops the Webots Y coordinate.

    .. deprecated:: 2.1.0
        Use ``WebotsCoordinateSystem`` instead.
    """
    x, y, z = pos
    return (x, z)


def scenicToWebotsPosition(pos, y=0, coordinateSystem="ENU"):
    """Convert a Scenic position to a Webots position.

    .. deprecated:: 2.1.0
        Use ``WebotsCoordinateSystem`` instead.
    """
    x, z = pos
    return [x, y, z]


def webotsToScenicRotation(rot, tolerance2D=None):
    """Convert a Webots rotation vector to a Scenic heading.
    Assumes the object lies in the Webots X-Z plane, with a rotation axis
    close to the Y axis. If ``tolerance2D`` is given, returns ``None`` if the
    orientation of the object is not sufficiently close to being 2D.

    .. deprecated:: 2.1.0
        Use ``WebotsCoordinateSystem`` instead.
    """
    axis = np.array(rot[:3])
    angle = rot[3]
    if tolerance2D is not None and np.linalg.norm(axis - (0, 1, 0)) > tolerance2D:
        return None
    return normalizeAngle(angle)


def scenicToWebotsRotation(heading):
    """Convert a Scenic heading to a Webots rotation vector.

    .. deprecated:: 2.1.0
        Use ``WebotsCoordinateSystem`` instead.
    """
    return [0, 1, 0, heading]



"""Extracting relations (for later pruning) from the syntax of requirements."""

from ast import (
    Add,
    BinOp,
    Call,
    Compare,
    Eq,
    Expression,
    Gt,
    GtE,
    Lt,
    LtE,
    Name,
    NotEq,
    Sub,
)
from collections import defaultdict
import math

from scenic.core.distributions import needsSampling
from scenic.core.errors import InconsistentScenarioError, InvalidScenarioError
from scenic.core.object_types import Object, Point


def inferRelationsFrom(reqNode, namespace, ego, line):
    """Infer relations between objects implied by a requirement."""
    matcher = RequirementMatcher(namespace)

    inferRelativeHeadingRelations(matcher, reqNode, ego, line)
    inferDistanceRelations(matcher, reqNode, ego, line)


def inferRelativeHeadingRelations(matcher, reqNode, ego, line):
    """Infer bounds on relative headings from a requirement."""
    rhMatcher = lambda node: matcher.matchUnaryFunction("RelativeHeading", node)
    allBounds = matcher.matchBounds(reqNode, rhMatcher)
    for target, bounds in allBounds:
        if not isinstance(target, Object):
            continue
        assert target is not ego
        if ego is None:
            raise InvalidScenarioError(
                "relative heading w.r.t. unassigned ego on line {line}"
            )
        lower, upper = bounds
        if lower < -math.pi:
            lower = -math.pi
        if upper > math.pi:
            upper = math.pi
        if lower == -math.pi and upper == math.pi:
            continue  # skip trivial bounds
        rel = RelativeHeadingRelation(target, lower, upper)
        ego._relations.append(rel)
        conv = RelativeHeadingRelation(ego, -upper, -lower)
        target._relations.append(conv)


def inferDistanceRelations(matcher, reqNode, ego, line):
    """Infer bounds on distances from a requirement."""
    distMatcher = lambda node: matcher.matchUnaryFunction("DistanceFrom", node)
    allBounds = matcher.matchBounds(reqNode, distMatcher)
    for target, bounds in allBounds:
        if not isinstance(target, Object):
            continue
        assert target is not ego
        if ego is None:
            raise InvalidScenarioError("distance w.r.t. unassigned ego on line {line}")
        lower, upper = bounds
        if lower < 0:
            lower = 0
            if upper == float("inf"):
                continue  # skip trivial bounds
        rel = DistanceRelation(target, lower, upper)
        ego._relations.append(rel)
        conv = DistanceRelation(ego, lower, upper)
        target._relations.append(conv)


class BoundRelation:
    """Abstract relation bounding something about another object."""

    def __init__(self, target, lower, upper):
        self.target = target
        self.lower, self.upper = lower, upper


class RelativeHeadingRelation(BoundRelation):
    """Relation bounding another object's relative heading with respect to this one."""

    pass


class DistanceRelation(BoundRelation):
    """Relation bounding another object's distance from this one."""

    pass


class RequirementMatcher:
    def __init__(self, namespace):
        self.namespace = namespace

    def inconsistencyError(self, node, message):
        raise InconsistentScenarioError(node.lineno, message)

    def matchUnaryFunction(self, name, node):
        """Match a call to a specified unary function, returning the value of its argument."""
        if not (
            isinstance(node, Call)
            and isinstance(node.func, Name)
            and node.func.id == name
        ):
            return None
        if len(node.args) != 1:
            return None
        if len(node.keywords) != 0:
            return None
        return self.matchValue(node.args[0])

    def matchBounds(self, node, matchAtom):
        """Match upper/lower bounds on something matched by the given function.

        Returns a list of all bounds found, pairing the bounded quantity with a
        pair (low, high) of lower/upper bounds.
        """
        if not isinstance(node, Compare):
            return {}
        bounds = defaultdict(lambda: (float("-inf"), float("inf")))
        targets = {}
        first = node.left
        for second, op in zip(node.comparators, node.ops):
            lower, upper, target = self.matchBoundsInner(first, second, op, matchAtom)
            first = second
            if target is None:
                continue
            targetID = id(target)  # use id to support unhashable types
            targets[targetID] = target
            bestLower, bestUpper = bounds[targetID]
            if lower is not None and lower > bestLower:
                bestLower = lower
            if upper is not None and upper < bestUpper:
                bestUpper = upper
            bounds[targetID] = (bestLower, bestUpper)
        return [(target, bounds[id_]) for id_, target in targets.items()]

    def matchBoundsInner(self, left, right, op, matchAtom):
        """Extract bounds from a single comparison operator."""
        # Reduce > and >= to < and <=
        if isinstance(op, Gt):
            return self.matchBoundsInner(right, left, Lt(), matchAtom)
        elif isinstance(op, GtE):
            return self.matchBoundsInner(right, left, LtE(), matchAtom)
        # Try matching a constant lower bound on the atom or its absolute value
        lconst = self.matchConstant(left)
        if isinstance(lconst, (int, float)):
            target = matchAtom(right)
            if target is not None:  # CONST op QUANTITY
                return (
                    (lconst, lconst, target)
                    if isinstance(op, Eq)
                    else (lconst, None, target)
                )
            else:
                bounds = self.matchAbsBounds(right, lconst, op, False, matchAtom)
                if bounds is not None:  # CONST op abs(QUANTITY [+/- CONST])
                    return bounds
        # Try matching a constant upper bound on the atom or its absolute value
        rconst = self.matchConstant(right)
        if isinstance(rconst, (int, float)):
            target = matchAtom(left)
            if target is not None:  # QUANTITY op CONST
                return (
                    (rconst, rconst, target)
                    if isinstance(op, Eq)
                    else (None, rconst, target)
                )
            else:
                bounds = self.matchAbsBounds(left, rconst, op, True, matchAtom)
                if bounds is not None:  # abs(QUANTITY [+/- CONST]) op CONST
                    return bounds
        return None, None, None

    def matchAbsBounds(self, node, const, op, isUpperBound, matchAtom):
        """Extract bounds on an atom from a comparison involving its absolute value."""
        if not (
            isinstance(node, Call)
            and isinstance(node.func, Name)
            and node.func.id == "abs"
        ):
            return None  # not an invocation of abs
        if not isUpperBound and not isinstance(op, Eq):
            return None  # lower bounds on abs value don't bound underlying quantity
        if const < 0:
            self.inconsistencyError(node, f"absolute value cannot be negative")
        assert len(node.args) == 1
        arg = node.args[0]
        target = matchAtom(arg)
        if target is not None:  # abs(QUANTITY) </= CONST
            return (-const, const, target)
        elif isinstance(arg, BinOp) and isinstance(arg.op, (Add, Sub)):
            # abs(X +/- Y) </= CONST
            match = None
            slconst = self.matchConstant(arg.left)
            target = matchAtom(arg.right)
            if isinstance(slconst, (int, float)) and target is not None:
                # abs(CONST +/- QUANTITY) </= CONST
                match = slconst
            else:
                srconst = self.matchConstant(arg.right)
                target = matchAtom(arg.left)
                if isinstance(srconst, (int, float)) and target is not None:
                    # abs(QUANTITY +/- CONST) </= CONST
                    match = srconst
            if match is not None:
                if isinstance(arg.op, Add):  # abs(QUANTITY + CONST) </= CONST
                    return (-const - match, const - match, target)
                else:  # abs(QUANTITY - CONST) </= CONST
                    return (-const + match, const + match, target)
        return None

    def matchConstant(self, node):
        """Match constant values, i.e. values known prior to sampling."""
        value = self.matchValue(node)
        return None if needsSampling(value) else value

    def matchValue(self, node):
        """Match any expression which can be evaluated, returning its value.

        This method could have undesirable side-effects, but conditions in
        requirements should not have side-effects to begin with.
        """
        try:
            code = compile(Expression(node), "<internal>", "eval")
            value = eval(code, dict(self.namespace))
        except Exception:
            return None
        return value



"""Domain for driving scenarios.

This domain must currently be used in `2D compatibility mode`.

The :doc:`world model <scenic.domains.driving.model>` defines Scenic classes for cars,
pedestrians, etc., actions for dynamic agents which walk or drive, as well as simple
behaviors like lane-following. Scenarios for the driving domain should import the model
as follows::

    model scenic.domains.driving.model

Scenarios written for the driving domain should work without changes [#f1]_ in any of the
following simulators:

    * CARLA, using the model :doc:`scenic.simulators.carla.model`
    * LGSVL, using the model :doc:`scenic.simulators.lgsvl.model`
    * the built-in Newtonian simulator, using the model
      :doc:`scenic.simulators.newtonian.driving_model`

For example, the :file:`examples/driving/badlyParkedCarPullingIn.scenic` scenario is
written for the driving domain and can be run in multiple simulators:

    * no simulator, for viewing the initial scene:

        .. code-block:: console

            $ scenic examples/driving/badlyParkedCarPullingIn.scenic

    * the built-in Newtonian simulator, for quick debugging without having to install an
      external simulator:

        .. code-block:: console

            $ scenic -S --model scenic.simulators.newtonian.driving_model \\
                examples/driving/badlyParkedCarPullingIn.scenic

    * CARLA, using the default map specified in the scenario:

        .. code-block:: console

            $ scenic -S --model scenic.simulators.carla.model \\
                examples/driving/badlyParkedCarPullingIn.scenic

    * LGSVL, specifying a map which it supports:

        .. code-block:: console

            $ scenic -S --model scenic.simulators.lgsvl.model \\
                --param map tests/formats/opendrive/maps/LGSVL/borregasave.xodr \\
                --param lgsvl_map BorregasAve \\
                examples/driving/badlyParkedCarPullingIn.scenic

.. rubric:: Footnotes

.. [#f1] Assuming the simulator supports the selected map. If necessary, the map may be
    changed from the command line using the :option:`--param` option; see the
    :doc:`model documentation <scenic.domains.driving.model>` for details.
"""



"""Support for values which are sampled outside of Scenic.

External Samplers in General
============================

External samplers provide a mechanism to use different types of sampling
techniques, like optimization or quasi-random sampling, from within a Scenic
program. Ordinary random values in Scenic are instances of `Distribution`;
this module defines a special subclass, `ExternalParameter`, representing a
value which is sampled externally. Scenic programs with external parameters
are handled as follows:

    1. During compilation, all instances of `ExternalParameter` are gathered
       together and given to the `ExternalSampler.forParameters` function;
       this function creates an appropriate `ExternalSampler`,
       whose configuration can be controlled using :term:`global parameters`
       (see the function documentation for details).

    2. When sampling a scene, before sampling any other distributions the
       :obj:`~ExternalSampler.sample` method of the `ExternalSampler` is
       called to sample all the external parameters. For active samplers, this
       method passes along the ``feedback`` value given to `Scenario.generate`,
       if any.

    3. Once the external parameters have values, the program is equivalent to
       one without external parameters, and sampling proceeds as usual. As for
       every instance of `Distribution`, the external parameters will have
       their :obj:`~Samplable.sampleGiven` method called once all their
       dependencies have been sampled; by default this method just returns the
       value sampled for this parameter in step (2).

.. note::

    Note that while external parameters, like all instances of `Distribution`,
    are allowed to have dependencies, they are an exception to the usual rule
    that dependencies are always sampled before dependents, because the
    `ExternalSampler.sample` method is called before any other sampling.
    However, as explained above, the :obj:`~Samplable.sampleGiven` method is
    called in the proper order and external samplers which need to do sampling
    based on the values of other distributions can be invoked from it. The
    two-step mechanism with `ExternalSampler.sample` is provided for samplers
    which sample the whole space of external parameters at once (e.g. the
    VerifAI samplers).

Samplers from VerifAI
=====================

The external sampling mechanism is designed to be extensible. The only built-in
`ExternalSampler` is the `VerifaiSampler`, which provides access to the
samplers in the `VerifAI`_ toolkit (which in turn can use Scenic as a modeling
language).

The `VerifaiSampler` supports several types of external parameters corresponding
to the primitive distributions: `VerifaiRange` and `VerifaiDiscreteRange` for
continuous and discrete intervals, and `VerifaiOptions` for discrete sets.
For example, suppose we write::

    ego = new Object at (VerifaiRange(5, 15), 0)

This is equivalent to the ordinary Scenic line :scenic:`ego = new Object at (Range(5, 15), 0)`,
except that the X coordinate of the ego is sampled by VerifAI within the range
(5, 15) instead of being uniformly distributed over it. By default the
`VerifaiSampler` uses VerifAI's `Halton`_ sampler, so the range will still be
covered uniformly but more systematically. If we want to use a different sampler,
we can set the ``verifaiSamplerType`` global parameter::

    param verifaiSamplerType = 'ce'
    ego = new Object at (VerifaiRange(5, 15), 0)

Now the X coordinate will be sampled using VerifAI's `cross-entropy`_ sampler.
If we pass a feedback value to `Scenario.generate` which scores the previous
scene, then the coordinate will not be sampled uniformly but rather converge to
a distribution concentrated on values minimizing the score. Active samplers like
cross-entropy can be used for falsification in this way, driving a system toward
parts of the parameter space where a specification is violated.

The cross-entropy sampler in VerifAI can be started from a non-uniform prior.
Scenic provides a convenient way to define this prior using the ordinary syntax
for distributions::

    param verifaiSamplerType = 'ce'
    ego = new Object at (VerifaiParameter.withPrior(Normal(10, 3)), 0)

Now cross-entropy sampling will start from a normal distribution with mean 10
and standard deviation 3. Priors are restricted to primitive distributions and
in general may be approximated so that VerifAI can handle them -- see
`VerifaiParameter.withPrior` for details.

For more information on how to customize the sampler, see `VerifaiSampler`.

.. _VerifAI: https://github.com/BerkeleyLearnVerify/VerifAI

.. _Halton: https://en.wikipedia.org/wiki/Halton_sequence

.. _cross-entropy: https://en.wikipedia.org/wiki/Cross-entropy_method

"""

from dotmap import DotMap
import numpy

from scenic.core.distributions import Distribution, Options
from scenic.core.errors import InvalidScenarioError


class ExternalSampler:
    """Abstract class for objects called to sample values for each external parameter.

    The initializer for this class takes the same arguments as the factory function
    `forParameters` below.

    Attributes:
        rejectionFeedback: Value passed to the `sample` method when the last sample was rejected.
          This value can be chosen by a Scenic scenario using the global parameter
          ``externalSamplerRejectionFeedback``.
    """

    def __init__(self, params, globalParams):
        # feedback value passed to external sampler when the last scene was rejected
        self.rejectionFeedback = globalParams.get("externalSamplerRejectionFeedback")

    @staticmethod
    def forParameters(params, globalParams):
        """Create an `ExternalSampler` given the sets of external and global parameters.

        The scenario may explicitly select an external sampler by assigning the
        :term:`global parameter` ``externalSampler`` to a subclass of `ExternalSampler`.
        Otherwise, a `VerifaiSampler` is used by default.

        Args:
            params (tuple): Tuple listing each `ExternalParameter`.
            globalParams (dict): Dictionary of global parameters for the `Scenario`, made
              available here to support sampler customization through setting parameters.
              Note that the values of these parameters may be instances of `Distribution`!

        Returns:
            An `ExternalSampler` configured for the given parameters.
        """
        if len(params) > 0:
            externalSampler = globalParams.get("externalSampler", VerifaiSampler)
            if not issubclass(externalSampler, ExternalSampler):
                raise InvalidScenarioError(
                    f"externalSampler type {externalSampler}"
                    " not subclass of ExternalSampler"
                )
            return externalSampler(params, globalParams)
        else:
            return None

    def sample(self, feedback):
        """Sample values for all the external parameters.

        Args:
            feedback: Feedback from the last sample (for active samplers).
        """
        self.cachedSample = self.nextSample(feedback)

    def nextSample(self, feedback):
        """Actually do the sampling. Implemented by subclasses."""
        raise NotImplementedError

    def valueFor(self, param):
        """Return the sampled value for a parameter. Implemented by subclasses."""
        raise NotImplementedError


class VerifaiSampler(ExternalSampler):
    """An external sampler exposing the samplers in the VerifAI toolkit.

    The sampler can be configured using the following Scenic :term:`global parameters`:

        * ``verifaiSamplerType`` -- sampler type (see the ``verifai.server.choose_sampler``
          function); the default is ``'halton'``
        * ``verifaiSamplerParams`` -- ``DotMap`` of options passed to the sampler

    The `VerifaiSampler` supports external parameters which are instances of `VerifaiParameter`.
    """

    def __init__(self, params, globalParams):
        super().__init__(params, globalParams)
        import verifai.features
        import verifai.server

        # construct FeatureSpace
        usingProbs = False
        self.params = tuple(params)
        for index, param in enumerate(self.params):
            if not isinstance(param, VerifaiParameter):
                raise RuntimeError(
                    f"VerifaiSampler given parameter of wrong type: {param}"
                )
            param.sampler = self
            param.index = index
            if param.probs is not None:
                usingProbs = True
        space = verifai.features.FeatureSpace(
            {
                f"param{index}": verifai.features.Feature(param.domain)
                for index, param in enumerate(self.params)
            }
        )

        # set up VerifAI sampler
        samplerType = globalParams.get("verifaiSamplerType", "halton")
        samplerParams = globalParams.get("verifaiSamplerParams", None)
        if usingProbs and samplerType == "ce":
            if samplerParams is None:
                samplerParams = DotMap()
            else:
                samplerParams = samplerParams.copy()  # avoid mutating original
            if "cont" in samplerParams or "disc" in samplerParams:
                raise RuntimeError(
                    "CE distributions specified in both VerifaiParameters"
                    " and verifaiSamplerParams"
                )
            cont_buckets = []
            cont_dists = []
            disc_dists = []
            for param in self.params:
                if isinstance(param, VerifaiRange):
                    if param.probs is None:
                        buckets = 5
                        dist = numpy.ones(buckets) / buckets
                    else:
                        dist = numpy.array(param.probs)
                        buckets = len(dist)
                    cont_buckets.append(buckets)
                    cont_dists.append(dist)
                elif isinstance(param, VerifaiDiscreteRange):
                    n = param.high - param.low + 1
                    dist = (
                        numpy.ones(n) / n
                        if param.probs is None
                        else numpy.array(param.probs)
                    )
                    disc_dists.append(dist)
                else:
                    raise RuntimeError(f"Parameter {param} not supported by CE sampler")
            samplerParams.cont.buckets = cont_buckets
            samplerParams.cont.dist = numpy.array(cont_dists)
            samplerParams.disc.dist = numpy.array(disc_dists)
        data = verifai.server.choose_sampler(
            space, samplerType, sampler_params=samplerParams
        )
        if not data:
            raise RuntimeError(f'Unknown VerifAI sampler type "{samplerType}"')
        self.sampler = data[1]

        # default rejection feedback is positive so cross-entropy sampler won't update;
        # for other active samplers an appropriate value should be set manually
        if self.rejectionFeedback is None:
            self.rejectionFeedback = 1
        self.cachedSample = None

    def nextSample(self, feedback):
        return self.sampler.nextSample(feedback)

    def update(self, sample, info, rho):
        self.sampler.update(sample, info, rho)

    def getSample(self):
        return self.sampler.getSample()

    def valueFor(self, param):
        return self.cachedSample[param.index]


class ExternalParameter(Distribution):
    """A value determined by external code rather than Scenic's internal sampler."""

    def __init__(self):
        super().__init__()
        self.sampler = None
        import scenic.syntax.veneer as veneer  # TODO improve?

        veneer.registerExternalParameter(self)

    def sampleGiven(self, value):
        """Specialization of  `Samplable.sampleGiven` for external parameters.

        By default, this method simply looks up the value previously sampled by
        `ExternalSampler.sample`.
        """
        assert self.sampler is not None
        return self.sampler.valueFor(self)


class VerifaiParameter(ExternalParameter):
    """An external parameter sampled using one of VerifAI's samplers."""

    def __init__(self, domain):
        super().__init__()
        self.domain = domain

    @staticmethod
    def withPrior(dist, buckets=None):
        """Creates a `VerifaiParameter` using the given distribution as a prior.

        Since the VerifAI cross-entropy sampler currently only supports piecewise-constant
        distributions, if the prior is not of that form it may be approximated. For most
        built-in distributions, the approximation is exact: for a particular distribution,
        check its `bucket` method.
        """
        if not dist.isPrimitive:
            raise RuntimeError(
                "VerifaiParameter.withPrior called on "
                f"non-primitive distribution {dist}"
            )
        bucketed = dist.bucket(buckets=buckets)
        return VerifaiOptions(
            bucketed.optWeights if bucketed.optWeights else bucketed.options
        )


class VerifaiRange(VerifaiParameter):
    """A :obj:`~scenic.core.distributions.Range` (real interval) sampled by VerifAI."""

    _defaultValueType = float

    def __init__(self, low, high, buckets=None, weights=None):
        import verifai.features

        super().__init__(verifai.features.Box([low, high]))
        if weights is not None:
            weights = tuple(weights)
            if buckets is not None and len(weights) != buckets:
                raise RuntimeError(
                    f"VerifaiRange created with {len(weights)} weights "
                    f"but {buckets} buckets"
                )
        elif buckets is not None:
            weights = [1] * buckets
        else:
            self.probs = None
            return
        total = sum(weights)
        self.probs = tuple(wt / total for wt in weights)

    def sampleGiven(self, value):
        value = super().sampleGiven(value)
        assert len(value) == 1
        return value[0]


class VerifaiDiscreteRange(VerifaiParameter):
    """A :obj:`~scenic.core.distributions.DiscreteRange` (integer interval) sampled by VerifAI."""

    _defaultValueType = float

    def __init__(self, low, high, weights=None):
        import verifai.features

        super().__init__(verifai.features.DiscreteBox([low, high]))
        if weights is not None:
            if len(weights) != (high - low + 1):
                raise RuntimeError(
                    f"VerifaiDiscreteRange created with {len(weights)} weights "
                    f"for {high - low + 1} values"
                )
            total = sum(weights)
            self.probs = tuple(wt / total for wt in weights)
        else:
            self.probs = None

    def sampleGiven(self, value):
        value = super().sampleGiven(value)
        assert len(value) == 1
        return value[0]


class VerifaiOptions(Options):
    """An :obj:`~scenic.core.distributions.Options` (discrete set) sampled by VerifAI."""

    @staticmethod
    def makeSelector(n, weights):
        return VerifaiDiscreteRange(0, n, weights)



import itertools
import json
import math
import re

try:
    import pyproj
except ModuleNotFoundError as e:
    raise RuntimeError("guideways scenarios require pyproj to be installed") from e

import numpy as np

from scenic.core.geometry import headingOfSegment, polygonUnion
from scenic.core.regions import PolygonalRegion, PolylineRegion
from scenic.core.type_support import toVector
from scenic.core.vectors import VectorField
from scenic.core.workspaces import Workspace


def toWebots(point, proj):
    x, z = proj(*point)
    return (-x, z)


def localize(points, proj):
    return tuple(toWebots(pt, proj) for pt in points)


class Bordered:
    def __init__(self, proj, data):
        lefts = data["left_border"]
        rights = data["right_border"]
        points = lefts + list(reversed(rights))
        self.region = PolygonalRegion(localize(points, proj))

    def show(self, plt, style="r"):
        self.region.show(plt, style=style)


class Guideway(Bordered):
    def __init__(self, proj, data):
        super().__init__(proj, data)
        self.id = data["id"]
        self.direction = data["direction"]
        self.origin_lane = data["origin_lane"]["lane_id"]
        medianPoints = localize(data["median"], proj)
        self.median = PolylineRegion(medianPoints)
        self.medianPoints = np.array(medianPoints)
        self.region.orientation = VectorField(f"Road{self.id}Direction", self.directionAt)

    def directionAt(self, point):
        # TODO improve?
        median = self.medianPoints
        pt = np.array(point[:2])
        # find closest point on median
        dists = np.linalg.norm(median - pt, axis=1)
        i, dist = min(enumerate(dists), key=lambda p: p[1])
        # if at either end of median, use that direction
        if i == 0:
            x, y = median[:2]
        elif i == len(median) - 1:
            x, y = median[-2:]
        else:
            # otherwise, pick which of the two directions to use based on
            # which side of the angle bisector we are on
            d1, d2 = median[i + 1] - median[i], median[i - 1] - median[i]
            d1 /= np.linalg.norm(d1)
            d2 /= np.linalg.norm(d2)
            bisector = d1 + d2
            if pt[1] * bisector[0] - pt[0] * bisector[1] < 0:
                x, y = median[i - 1], median[i]
            else:
                x, y = median[i], median[i + 1]

        x = toVector(x)
        y = toVector(y)
        return headingOfSegment(x, y)


class Crosswalk(Bordered):
    def show(self, plt, style="b"):
        self.region.show(plt, style=style)


class ConflictZone:
    def __init__(self, inter, data):
        pts = localize(data["polygon"]["coordinates"][0], inter.projection)
        self.region = PolygonalRegion(pts)
        # TODO not working... some conflict zones have invalid IDs?
        # self.guideway1 = inter.guidewayWithID[data['guideway1_id']]
        # self.guideway2 = inter.guidewayWithID[data['guideway2_id']]

    def show(self, plt):
        self.region.show(plt, style="k")


def projectionAt(point):
    # taken from Webots OSM importer
    lon, lat = point
    utm_zone = 1 + math.floor((float(lon) + 180) / 6)
    hemisphere = "south" if lat < 0 else "north"
    string = (
        "+proj=ortho +%s +zone=%d +lon_0=%f +lat_0=%f +x_0=0 +y_0=0 +ellps=WGS84 +units=m +no_defs"
        % (hemisphere, utm_zone, lon, lat)
    )
    return pyproj.Proj(string)


class Intersection:
    def __init__(self, data, origin=None):
        if origin is None:
            origin = data["main_gw"]["median"][0]
        self.projection = projectionAt(origin)

        self.mainGuideway = Guideway(self.projection, data["main_gw"])
        self.vehicleGuideways = tuple(
            Guideway(self.projection, d) for d in data["vehicle_gw"]
        ) + (self.mainGuideway,)
        self.bikeGuideways = tuple(
            Guideway(self.projection, d) for d in data["bicycle_gw"]
        )
        self.railGuideways = tuple(Guideway(self.projection, d) for d in data["rail_gw"])
        guidewayWithID = {}
        for gw in itertools.chain(
            self.vehicleGuideways, self.bikeGuideways, self.railGuideways
        ):
            guidewayWithID[gw.id] = gw
        self.guidewayWithID = guidewayWithID

        self.crosswalks = tuple(Crosswalk(self.projection, d) for d in data["crosswalks"])
        self.conflictZones = tuple(ConflictZone(self, d) for d in data["conflict_zones"])

    def directionAt(self, point):
        for guideway in self.vehicleGuideways:
            if guideway.region.containsPoint(point):
                return guideway.directionAt(point)
        raise RuntimeError("tried to get road direction at point not in a guideway!")

    def show(self, plt):
        for guideway in self.vehicleGuideways:
            guideway.show(plt)
        self.mainGuideway.show(plt, style="g")
        for guideway in self.bikeGuideways:
            guideway.show(plt, style="c")
        for guideway in self.railGuideways:
            guideway.show(plt, style="m")
        for crosswalk in self.crosswalks:
            crosswalk.show(plt)
        for zone in self.conflictZones:
            zone.show(plt)


class IntersectionWorkspace(Workspace):
    def __init__(self, intersection):
        self.intersection = intersection
        drivablePoly = polygonUnion(
            gw.region.polygons for gw in intersection.vehicleGuideways
        )
        walkablePoly = polygonUnion(cw.region.polygons for cw in intersection.crosswalks)
        workspacePoly = polygonUnion((drivablePoly, walkablePoly))
        self.roadDirection = VectorField("RoadDirection", intersection.directionAt)
        self.drivableRegion = PolygonalRegion(
            polygon=drivablePoly, orientation=self.roadDirection
        )
        self.workspaceRegion = PolygonalRegion(polygon=workspacePoly)
        super().__init__(self.workspaceRegion)

    def show(self, plt):
        plt.gca().set_aspect("equal")
        self.intersection.show(plt)

    @property
    def minimumZoomSize(self):
        return 30



"""Translator turning Scenic programs into Scenario objects.

The top-level interface to Scenic is provided by two functions:

* `scenarioFromString` -- compile a string of Scenic code;
* `scenarioFromFile` -- compile a Scenic file.

These output a `Scenario` object, from which scenes can be generated.
See the documentation for `Scenario` for details.

When imported, this module hooks the Python import system in order to implement
the :keyword:`import` statement. This is only for the compiler's own use: it is
not allowed to import a Scenic module from Python, and attempting to do so will
fail with a `ModuleNotFoundError`.

Scenic is compiled in two main steps: translating the code into Python, and
executing the resulting Python module to generate a Scenario object encoding
the objects, distributions, etc. in the scenario. For details, see the function
`compileStream` below.
"""

import ast
import builtins
from contextlib import contextmanager
import dataclasses
import hashlib
import importlib
import importlib.abc
import importlib.util
import inspect
import io
import os
import sys
import time
import types
from typing import Optional

from scenic.core.distributions import RejectionException, toDistribution
from scenic.core.dynamics.scenarios import DynamicScenario
import scenic.core.errors as errors
from scenic.core.errors import InvalidScenarioError, PythonCompileError
from scenic.core.lazy_eval import needsLazyEvaluation
import scenic.core.pruning as pruning
from scenic.core.utils import cached_property
from scenic.syntax.compiler import compileScenicAST
from scenic.syntax.parser import parse_string
import scenic.syntax.veneer as veneer

### THE TOP LEVEL: compiling a Scenic program


@dataclasses.dataclass
class CompileOptions:
    """Internal class for capturing options used when compiling a scenario."""

    # N.B. update `hash` below when adding a new field

    #: Whether or not the scenario uses `2D compatibility mode`.
    mode2D: bool = False
    #: Overridden world model, if any.
    modelOverride: Optional[str] = None
    #: Overridden global parameters.
    paramOverrides: dict = dataclasses.field(default_factory=dict)
    #: Selected modular scenario, if any.
    scenario: Optional[str] = None

    @cached_property
    def hash(self):
        """Deterministic hash saved in serialized scenes to catch option mismatches."""
        stream = io.BytesIO()
        stream.write(bytes([self.mode2D]))
        if self.modelOverride:
            stream.write(self.modelOverride.encode())
        for key in sorted(self.paramOverrides.keys()):
            stream.write(key.encode())
            value = self.paramOverrides[key]
            if isinstance(value, (int, float, str)):
                stream.write(str(value).encode())
            else:
                stream.write([0])
        if self.scenario:
            stream.write(self.scenario.encode())
        # We can't use `hash` because it is not deterministic
        # (e.g. the hashes of strings are randomized)
        return hashlib.blake2b(stream.getvalue(), digest_size=4).digest()


def scenarioFromString(
    string,
    params={},
    model=None,
    scenario=None,
    *,
    filename="<string>",
    mode2D=False,
    **kwargs,
):
    """Compile a string of Scenic code into a `Scenario`.

    The optional **filename** is used for error messages.
    Other arguments are as in `scenarioFromFile`.
    """
    stream = io.BytesIO(string.encode())
    options = CompileOptions(modelOverride=model, paramOverrides=params, mode2D=mode2D)
    return _scenarioFromStream(stream, options, filename, scenario=scenario, **kwargs)


def scenarioFromFile(
    path, params={}, model=None, scenario=None, *, mode2D=False, **kwargs
):
    """Compile a Scenic file into a `Scenario`.

    Args:
        path (str): Path to a Scenic file.
        params (dict): :term:`Global parameters` to override, as a dictionary mapping
          parameter names to their desired values.
        model (str): Scenic module to use as :term:`world model`.
        scenario (str): If there are multiple :term:`modular scenarios` in the
          file, which one to compile; if not specified, a scenario called 'Main'
          is used if it exists.
        mode2D (bool): Whether to compile this scenario in `2D compatibility mode`.

    Returns:
        A `Scenario` object representing the Scenic scenario.

    Note for Scenic developers: this function accepts additional keyword
    arguments which are intended for internal use and debugging only.
    See `_scenarioFromStream` for details.
    """
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    fullpath = os.path.realpath(path)
    head, extension = os.path.splitext(fullpath)
    if not extension or extension not in scenicExtensions:
        ok = ", ".join(scenicExtensions)
        err = f"Scenic scenario does not have valid extension ({ok})"
        raise RuntimeError(err)
    directory, name = os.path.split(head)

    options = CompileOptions(modelOverride=model, paramOverrides=params, mode2D=mode2D)
    with open(path, "rb") as stream:
        return _scenarioFromStream(
            stream, options, fullpath, scenario=scenario, path=path, **kwargs
        )


def _scenarioFromStream(
    stream, compileOptions, filename, *, scenario=None, path=None, _cacheImports=False
):
    """Compile a stream of Scenic code into a `Scenario`.

    This method is not meant to be called directly by users of Scenic. Use the
    top-level functions `scenarioFromFile` and `scenarioFromString` instead.

    These functions also accept the following keyword arguments, which are
    intended for internal use and debugging only. They should be considered
    unstable and are subject to modification or removal at any time.

    Args:
        _cacheImports (bool): Whether to cache any imported Scenic modules.
          The default behavior is to not do this, so that subsequent attempts
          to import such modules will cause them to be recompiled. If it is
          safe to cache Scenic modules across multiple compilations, set this
          argument to True. Then importing a Scenic module will have the same
          behavior as importing a Python module. See `purgeModulesUnsafeToCache`
          for a more detailed discussion of the internals behind this.
    """
    # Compile the code as if it were a top-level module
    oldModules = list(sys.modules.keys())
    try:
        with topLevelNamespace(path) as namespace:
            compileStream(stream, namespace, compileOptions, filename)
    finally:
        if not _cacheImports:
            purgeModulesUnsafeToCache(oldModules)
    # Construct a Scenario from the resulting namespace
    return constructScenarioFrom(namespace, scenario)


@contextmanager
def topLevelNamespace(path=None):
    """Creates an environment like that of a Python script being run directly.

    Specifically, __name__ is '__main__', __file__ is the path used to invoke
    the script (not necessarily its absolute path), and the parent directory is
    added to the path so that 'import blobbo' will import blobbo from that
    directory if it exists there.
    """
    directory = os.getcwd() if path is None else os.path.dirname(path)
    namespace = {"__name__": "__main__"}
    if path is not None:
        namespace["__file__"] = path
    sys.path.insert(0, directory)
    try:
        yield namespace
    finally:
        # Remove directory from sys.path, being a little careful in case the
        # Scenic program modified it (unlikely but possible).
        try:
            sys.path.remove(directory)
        except ValueError:
            pass


def purgeModulesUnsafeToCache(oldModules):
    """Uncache loaded modules which should not be kept after compilation.

    Keeping Scenic modules in `sys.modules` after compilation will cause
    subsequent attempts at compiling the same module to reuse the compiled
    scenario: this is usually not what is desired, since compilation can depend
    on external state (in particular overridden global parameters, used e.g. to
    specify the map for driving domain scenarios).

    Args:
        oldModules: List of names of modules loaded before compilation. These
            will be skipped.
    """
    toRemove = []
    # copy sys.modules in case it mutates during iteration (actually happens!)
    for name, module in sys.modules.copy().items():
        if isinstance(module, ScenicModule) and name not in oldModules:
            toRemove.append(name)
    for name in toRemove:
        parent, _, child = name.rpartition(".")
        parent = sys.modules.get(parent)
        if parent:
            # Remove reference to purged module from parent module. This is necessary
            # so that future imports of the purged module will properly refer to the
            # newly-loaded version of it. See below for a long disquisition on this.
            del parent.__dict__[child]

            # Here are details on why the above line is necessary and the sorry history
            # of my attempts to fix this type of bug (hopefully this note will prevent
            # further self-sabotage). Suppose we have a Python package 'package'
            # with a Scenic submodule 'submodule'. A Scenic program with the line
            #   from package import submodule
            # will import 2 packages, namely package and package.submodule, when first
            # compiled. We will then purge package.submodule from sys.modules, but not
            # package, since it is an ordinary module. So if the program is compiled a
            # second time, the line above will NOT import package.submodule, but simply
            # access the attribute 'submodule' of the existing package 'package'. So the
            # reference to the old version of package.submodule will leak out.
            # (An alternative approach, which I used to use, would be to purge all
            # modules containing even indirect references to Scenic modules, but this
            # opens a can of worms: the implementation of
            #   import parent.child
            # does not set the 'child' attribute of 'parent' if 'parent.child' is already
            # in sys.modules, violating an invariant that Python expects [see
            # https://docs.python.org/3/reference/import.html#submodules] and leading to
            # confusing errors. So if parent is purged because it has some child which is
            # a Scenic module, *all* of its children must then be purged. Since the
            # scenic module itself can contain indirect references to Scenic modules (the
            # world models), this means we have to purge the entire scenic package. But
            # then whoever did 'import scenic' at the top level will be left with a
            # reference to the old version of the Scenic module.)
            #
            # 2023 update: after hitting yet another bug caused by a reference to a
            # Scenic module surviving the purge, I've decided to completely ban importing
            # Scenic modules from Python (except when building the documentation). Any
            # objects needed in both Python and Scenic modules should be defined in a
            # Python module and imported from there.
        del sys.modules[name]


def compileStream(stream, namespace, compileOptions, filename):
    """Compile a stream of Scenic code and execute it in a namespace.

    The compilation procedure consists of the following main steps:

        1. Parse the Scenic code into a Scenic AST using the parser generated
           by ``pegen`` from :file:`scenic.gram`.
        2. Compile the Scenic AST into a Python AST with the desired semantics.
           This is done by the compiler, `scenic.syntax.compiler`.
        3. Compile and execute the Python AST.
        4. Extract the global state (e.g. objects).
           This is done by the `storeScenarioStateIn` function.
    """
    if errors.verbosityLevel >= 2:
        veneer.verbosePrint(f"  Compiling Scenic module from {filename}...")
        startTime = time.time()
    veneer.activate(compileOptions, namespace)
    try:
        # Execute preamble
        exec(compile(preamble, "<veneer>", "exec"), namespace)
        namespace[namespaceReference] = namespace

        # Parse the source
        source = stream.read().decode("utf-8")
        scenic_tree = parse_string(source, "exec", filename=filename)

        if dumpScenicAST:
            print(f"### Begin Scenic AST of {filename}")
            print(dump(scenic_tree, include_attributes=False, indent=4))
            print("### End Scenic AST")

        # Compile the Scenic AST into a Python AST
        tree, requirements = compileScenicAST(scenic_tree, filename=filename)
        astHasher = hashlib.blake2b(digest_size=4)
        astHasher.update(ast.dump(tree).encode())

        if dumpFinalAST:
            print(f"### Begin final AST of {filename}")
            print(dump(tree, include_attributes=True, indent=4))
            print("### End final AST")

        pythonSource = astToSource(tree)
        if dumpASTPython:
            if pythonSource is None:
                raise RuntimeError(
                    "dumping the Python equivalent of the AST"
                    " requires the astor package"
                )
            print(f"### Begin Python equivalent of final AST of {filename}")
            print(pythonSource)
            print("### End Python equivalent of final AST")

        # Compile the Python AST tree
        code = compileTranslatedTree(tree, filename)

        # Execute it
        executeCodeIn(code, namespace)

        # Extract scenario state from veneer and store it
        astHash = astHasher.digest()
        storeScenarioStateIn(namespace, requirements, astHash, compileOptions)
    finally:
        veneer.deactivate()
    if errors.verbosityLevel >= 2:
        totalTime = time.time() - startTime
        veneer.verbosePrint(f"  Compiled Scenic module in {totalTime:.4g} seconds.")
    return code, pythonSource


def dump(
    node: ast.AST,
    annotate_fields: bool = True,
    include_attributes: bool = False,
    *,
    indent: int,
):
    if sys.version_info >= (3, 9):
        print(ast.dump(node, annotate_fields, include_attributes, indent=indent))
    else:
        # omit `indent` if not supported
        print(ast.dump(node, annotate_fields, include_attributes))


def astToSource(tree: ast.AST):
    if sys.version_info >= (3, 9):
        return ast.unparse(tree)
    try:
        import astor
    except ModuleNotFoundError:
        return None
    return astor.to_source(tree)


### TRANSLATION PHASE ZERO: definitions of language elements not already in Python

## Options

dumpScenicAST = False
dumpFinalAST = False
dumpASTPython = False
usePruning = True

## Preamble
# (included at the beginning of every module to be translated;
# imports the implementations of the public language features)
preamble = """\
from scenic.syntax.veneer import *
"""

## Get Python names of various elements
## (for checking consistency between the translator and the veneer)

api = set(veneer.__all__)

namespaceReference = "_Scenic_module_namespace"  # used in the implementation of 'model'

### TRANSLATION PHASE ONE: handling imports

## Loader for Scenic files, producing ScenicModules


class ScenicLoader(importlib.abc.InspectLoader):
    def __init__(self, name, filepath):
        self.filepath = filepath

    def create_module(self, spec):
        return ScenicModule(spec.name)

    def exec_module(self, module):
        # Read source file and compile it
        with open(self.filepath, "r") as stream:
            source = stream.read()
        with open(self.filepath, "rb") as stream:
            code, pythonSource = compileStream(
                stream, module.__dict__, CompileOptions(), self.filepath
            )
        # Save code, source, and translated source for later inspection
        module._code = code
        module._source = source
        module._pythonSource = pythonSource

        # If we're in the process of compiling another Scenic module, inherit
        # objects, parameters, etc. from this one
        if veneer.isActive():
            veneer.currentScenario._inherit(module._scenario)

    def is_package(self, fullname):
        return False

    def get_code(self, fullname):
        module = importlib.import_module(fullname)
        assert isinstance(module, ScenicModule), module
        return module._code

    def get_source(self, fullname):
        module = importlib.import_module(fullname)
        assert isinstance(module, ScenicModule), module
        return module._source


class ScenicModule(types.ModuleType):
    def __getstate__(self):
        state = self.__dict__.copy()
        del state["__builtins__"]
        return (self.__name__, state)

    def __setstate__(self, state):
        name, state = state
        self.__init__(name)  # needed to create __dict__
        self.__dict__.update(state)
        self.__builtins__ = builtins.__dict__


# Give instances of ScenicModule a falsy __module__ to prevent Sphinx from
# getting confused. (Autodoc doesn't expect modules to have that attribute,
# and we can't del it.) We only do this during Sphinx runs since it seems to
# sometimes break pickling of the modules.
sphinx = sys.modules.get("sphinx")
buildingDocs = sphinx and getattr(sphinx, "_buildingScenicDocs", False)
if buildingDocs:
    ScenicModule.__module__ = None

## Finder for Scenic (and Python) files

scenicExtensions = (".scenic", ".sc")

import importlib.machinery as machinery

loaders = [
    (machinery.ExtensionFileLoader, machinery.EXTENSION_SUFFIXES),
    (machinery.SourceFileLoader, machinery.SOURCE_SUFFIXES),
    (machinery.SourcelessFileLoader, machinery.BYTECODE_SUFFIXES),
    (ScenicLoader, scenicExtensions),
]


class ScenicFileFinder(importlib.abc.PathEntryFinder):
    def __init__(self, path):
        self._inner = machinery.FileFinder(path, *loaders)

    def find_spec(self, fullname, target=None):
        spec = self._inner.find_spec(fullname, target=target)
        # Disallow imports of Scenic modules from Python modules, unless we are
        # building the documentation (to allow autodoc to introspect them; this
        # requires careful setup in `docs/conf.py`).
        # See `purgeModulesUnsafeToCache` for the rationale.
        if (
            spec
            and spec.origin
            and not (veneer.isActive() or buildingDocs)
            and any(spec.origin.endswith(ext) for ext in scenicExtensions)
        ):
            return None
        return spec

    def invalidate_caches(self):
        self._inner.invalidate_caches()

    # Support pkgutil.iter_modules() (used by Sphinx autosummary's recursive mode);
    # we need to use a subclass of FileFinder since pkgutil's implementation for
    # vanilla FileFinder uses inspect.getmodulename, which doesn't recognize the
    # .scenic file extension.
    def iter_modules(self, prefix):
        # This is mostly copied from pkgutil._iter_file_finder_modules
        yielded = {}
        try:
            filenames = os.listdir(self._inner.path)
        except OSError:
            return
        filenames.sort()
        for fn in filenames:
            modname = inspect.getmodulename(fn)
            if not modname:
                # Check for Scenic modules
                base = os.path.basename(fn)
                for ext in scenicExtensions:
                    if base.endswith(ext):
                        modname = base[: -len(ext)]
                        break
            if modname == "__init__" or modname in yielded:
                continue

            path = os.path.join(self._inner.path, fn)
            ispkg = False

            if not modname and os.path.isdir(path) and "." not in fn:
                modname = fn
                try:
                    dircontents = os.listdir(path)
                except OSError:
                    # ignore unreadable directories like import does
                    dircontents = []
                for fn in dircontents:
                    subname = inspect.getmodulename(fn)
                    if subname == "__init__":
                        ispkg = True
                        break
                else:
                    continue  # not a package

            if modname and "." not in modname:
                yielded[modname] = 1
                yield prefix + modname, ispkg


# Install path hook using our finder
def scenic_path_hook(path):
    if not path:
        path = os.getcwd()
    if not os.path.isdir(path):
        raise ImportError("only directories are supported", path=path)
    return ScenicFileFinder(path)


sys.path_hooks.insert(0, scenic_path_hook)
sys.path_importer_cache.clear()

### Translation phase two to four are done by the parser & compiler

### TRANSLATION PHASE FIVE: AST compilation


def compileTranslatedTree(tree, filename):
    try:
        return compile(tree, filename, "exec")
    except SyntaxError as e:
        raise PythonCompileError(e) from None


### TRANSLATION PHASE SIX: Python execution


def executeCodeIn(code, namespace):
    """Execute the final translated Python code in the given namespace."""
    try:
        exec(code, namespace)
    except RejectionException as e:
        # Determined statically that the scenario has probability zero.
        errors.optionallyDebugRejection(e)
        if errors.showInternalBacktrace:
            raise InvalidScenarioError(e.args[0]) from e
        else:
            raise InvalidScenarioError(e.args[0]).with_traceback(
                e.__traceback__
            ) from None


### TRANSLATION PHASE SEVEN: scenario construction


def storeScenarioStateIn(namespace, requirementSyntax, astHash, options):
    """Post-process an executed Scenic module, extracting state from the veneer."""

    # Save requirement syntax and other module-level information
    namespace["_astHash"] = astHash
    namespace["_compileOptions"] = options
    moduleScenario = veneer.currentScenario
    factory = veneer.simulatorFactory
    bns = gatherBehaviorNamespacesFrom(moduleScenario._behaviors)

    def handle(scenario):
        scenario._requirementSyntax = requirementSyntax
        if isinstance(scenario, type):
            scenario._simulatorFactory = staticmethod(factory)
        else:
            scenario._simulatorFactory = factory
        scenario._behaviorNamespaces = bns

    handle(moduleScenario)
    namespace["_scenarios"] = tuple(veneer.scenarios)
    for scenarioClass in veneer.scenarios:
        handle(scenarioClass)

    # Extract requirements, scan for relations used for pruning, and create closures
    # (only for top-level scenario; modular scenarios will be handled when instantiated)
    moduleScenario._compileRequirements()

    # Save global parameters
    for name, value in veneer._globalParameters.items():
        if needsLazyEvaluation(value):
            raise InvalidScenarioError(
                f"parameter {name} uses value {value}"
                " undefined outside of object definition"
            )
    for scenario in veneer.scenarios:
        scenario._bindGlobals(veneer._globalParameters)
    moduleScenario._bindGlobals(veneer._globalParameters)

    namespace["_scenario"] = moduleScenario


def gatherBehaviorNamespacesFrom(behaviors):
    """Gather any global namespaces which could be referred to by behaviors.

    We'll need to rebind any sampled values in them at runtime.
    """
    behaviorNamespaces = {}

    def registerNamespace(modName, ns):
        oldNS = behaviorNamespaces.get(modName)
        if oldNS:
            # Already registered; just do a consistency check to avoid bizarre
            # bugs from having multiple versions of the same module around.
            if oldNS is not ns:
                raise RuntimeError(
                    f"scenario refers to multiple versions of module {modName}; "
                    "perhaps you imported it before you started compilation?"
                )
            return
        behaviorNamespaces[modName] = ns
        for name, value in ns.items():
            if isinstance(value, ScenicModule):
                registerNamespace(value.__name__, value.__dict__)
            else:
                # Convert values requiring sampling to Distributions
                dval = toDistribution(value)
                if dval is not value:
                    ns[name] = dval

    for behavior in behaviors:
        modName = behavior.__module__
        globalNamespace = behavior.makeGenerator.__globals__
        registerNamespace(modName, globalNamespace)
    return behaviorNamespaces


def constructScenarioFrom(namespace, scenarioName=None):
    """Build a Scenario object from an executed Scenic module."""
    modularScenarios = namespace["_scenarios"]

    def isModularScenario(thing):
        return isinstance(thing, type) and issubclass(thing, DynamicScenario)

    if not scenarioName and isModularScenario(namespace.get("Main", None)):
        scenarioName = "Main"
    if scenarioName:
        ty = namespace.get(scenarioName, None)
        if not isModularScenario(ty):
            raise RuntimeError(f'no scenario "{scenarioName}" found')
        if ty._requiresArguments():
            raise RuntimeError(
                f'cannot instantiate scenario "{scenarioName}"' " with no arguments"
            ) from None

        dynScenario = ty()
    elif len(modularScenarios) > 1:
        raise RuntimeError(
            "multiple choices for scenario to run "
            "(specify using the --scenario option)"
        )
    elif modularScenarios and not modularScenarios[0]._requiresArguments():
        dynScenario = modularScenarios[0]()
    else:
        dynScenario = namespace["_scenario"]

    if not dynScenario._prepared:  # true for all except top-level scenarios
        # Execute setup block (if any) to create objects and requirements;
        # extract any requirements and scan for relations used for pruning
        dynScenario._prepare(delayPreconditionCheck=True)
    scenario = dynScenario._toScenario(namespace)

    # Prune infeasible parts of the space
    if usePruning:
        pruning.prune(scenario, verbosity=errors.verbosityLevel)

    # Validate scenario
    scenario.validate()

    return scenario
